"""This module implements error handlers for Gaussian runs."""

from __future__ import annotations

import datetime
import glob
import logging
import math
import os
import re
import shutil
from typing import TYPE_CHECKING, Any, ClassVar

import numpy as np
from monty.io import zopen
from pymatgen.io.gaussian import GaussianInput, GaussianOutput

from custodian.custodian import ErrorHandler
from custodian.utils import backup

if TYPE_CHECKING:
    from collections.abc import Iterable

__author__ = "Rasha Atwi"
__version__ = "0.1"
__maintainer__ = "Rasha Atwi"
__email__ = "rasha.atwi@stonybrook.edu"
__status__ = "Alpha"
__date__ = "5/13/21"

BACKUP_FILES = {
    "checkpoint": "*.[Cc][Hh][Kk]",
    "form_checkpoint": "*.[Ff][Cc][Hh][Kk]",
    "rwf": "*.[Rr][Ww][Ff]",
    "inp": "*.[Ii][Nn][Pp]",
    "int": "*.[Ii][Nn][Tt]",
    "d2e": "*.[Dd]2[Ee]",
    "skr": "*.[Ss][Kk][Rr]",
    "convergence": "convergence.png",
}


class GaussianErrorHandler(ErrorHandler):
    """
    Master GaussianErrorHandler class that handles a number of common errors that occur
    during Gaussian runs.
    """

    # definition of job errors as they appear in Gaussian output file
    error_defs: ClassVar = {
        "Optimization stopped": "opt_steps",
        "Convergence failure": "scf_convergence",
        "FormBX had a problem": "linear_bend",
        "Linear angle in Tors.": "linear_bend",
        "Inv3 failed in PCMMkU": "solute_solvent_surface",
        "Error in internal coordinate system": "internal_coords",
        "End of file in ZSymb": "zmatrix",
        "There are no atoms in this input structure !": "missing_mol",
        "Atom specifications unexpectedly found in input stream.": "found_coords",
        "End of file reading connectivity.": "coords",
        "FileIO operation on non-existent file.": "missing_file",
        "No data on chk file.": "empty_file",
        "Bad file opened by FileIO": "bad_file",
        "Z-matrix optimization but no Z-matrix variables.": "coord_inputs",
        "A syntax error was detected in the input line.": "syntax",
        r"The combination of multiplicity ([0-9]+) and \s+? ([0-9]+) electrons is impossible.": "charge",
        "Out-of-memory error in routine": "insufficient_mem",
    }

    error_patt = re.compile("|".join(list(error_defs)))
    recom_mem_patt = re.compile(
        r"Use %mem=([0-9]+)MW to provide the minimum amount of memory required to complete this step."
    )
    conv_criteria: ClassVar = {
        "max_force": re.compile(r"\s+(Maximum Force)\s+(-?\d+.?\d*|.*)\s+(-?\d+.?\d*)"),
        "rms_force": re.compile(r"\s+(RMS {5}Force)\s+(-?\d+.?\d*|.*)\s+(-?\d+.?\d*)"),
        "max_disp": re.compile(r"\s+(Maximum Displacement)\s+(-?\d+.?\d*|.*)\s+(-?\d+.?\d*)"),
        "rms_disp": re.compile(r"\s+(RMS {5}Displacement)\s+(-?\d+.?\d*|.*)\s+(-?\d+.?\d*)"),
    }

    grid_patt = re.compile(r"(-?\d{5})")
    GRID_NAMES = (
        "finegrid",
        "fine",
        "superfinegrid",
        "superfine",
        "coarsegrid",
        "coarse",
        "sg1grid",
        "sg1",
        "pass0grid",
        "pass0",
    )
    MEM_UNITS = ("kb", "mb", "gb", "tb", "kw", "mw", "gw", "tw")

    activate_better_guess = False

    def __init__(
        self,
        input_file: str,
        output_file: str,
        stderr_file: str = "stderr.txt",
        cart_coords: bool = True,
        scf_max_cycles: int = 100,
        opt_max_cycles: int = 100,
        job_type: str = "normal",
        lower_functional: str | None = None,
        lower_basis_set: str | None = None,
        prefix: str = "error",
        check_convergence: bool = True,
    ):
        """
        Initialize the GaussianErrorHandler class.

        Args:
            input_file (str): The name of the input file for the Gaussian job.
            output_file (str): The name of the output file for the Gaussian job.
            stderr_file (str): The name of the standard error file for the Gaussian job.
                Defaults to 'stderr.txt'.
            cart_coords (bool): Whether the coordinates are in cartesian format.
                Defaults to True.
            scf_max_cycles (int): The maximum number of SCF cycles. Defaults to 100.
            opt_max_cycles (int): The maximum number of optimization cycles. Defaults to
                100.
            job_type (str): The type of job to run. Supported options are 'normal' and
                'better_guess'. Defaults to 'normal'. If 'better_guess' is chosen, the
                job will be rerun at a lower level of theory to get a better initial
                guess of molecular orbitals or geometry, if needed.
            lower_functional (str): The lower level of theory to use for a better guess.
            lower_basis_set (str): The lower basis set to use for a better guess.
            prefix (str): The prefix to use for the backup files. Defaults to error,
                which means a series of error.1.tar.gz, error.2.tar.gz, ... will be
                generated.
            check_convergence (bool): Whether to check for convergence during an
                optimization job. Defaults to True. If True, the convergence data will
                be monitored and plotted (convergence criteria versus cycle number) and
                saved to a file called 'convergence.png'.
        """
        self.input_file = input_file
        self.output_file = output_file
        self.stderr_file = stderr_file
        self.cart_coords = cart_coords
        self.errors: set[str] = set()
        self.scf_max_cycles = scf_max_cycles
        self.opt_max_cycles = opt_max_cycles
        self.job_type = job_type
        self.lower_functional = lower_functional
        self.lower_basis_set = lower_basis_set
        self.prefix = prefix
        self.check_convergence = check_convergence
        self.conv_data: dict[str, dict[str, Any]] = {}
        self.recom_mem: float | None = None
        self.logger: logging.Logger = logging.getLogger(self.__class__.__name__)
        logging.basicConfig(level=logging.INFO)

    @staticmethod
    def _recursive_lowercase(obj: dict[str, Any] | str | Iterable[Any]) -> dict[str, Any] | str | Iterable[Any]:
        """
        Recursively convert all string elements in a given object to lowercase.

        This method iterates over the input object. If the object is a dictionary, it
        converts all its string keys and values to lowercase, applying the same logic
        recursively to the values. If the object is a string, it directly converts it
        to lowercase. If the object is iterable (but not a string or dictionary), it
        applies the same lowercase conversion to each element in the iterable. For all
        other types, the object is returned unchanged.

        Args:
            obj (dict | str | iterable): The object to be converted to lowercase.
                This can be a dictionary, a string, or any iterable collection.
                Non-iterable objects or non-string elements within iterables are
                returned unchanged.

        Returns:
            dict | str | iterable:
                A new object with all string elements converted to
                lowercase. The type of the returned object matches the type of the
                input `obj`.
        """
        if isinstance(obj, dict):
            return {k.lower(): GaussianErrorHandler._recursive_lowercase(v) for k, v in obj.items()}
        if isinstance(obj, str):
            return obj.lower()
        if hasattr(obj, "__iter__"):
            return [GaussianErrorHandler._recursive_lowercase(i) for i in obj]
        return obj

    @staticmethod
    def _recursive_remove_space(obj: dict[str, Any]) -> dict[str, Any]:
        """
        Recursively remove leading and trailing whitespace from keys and string values
        in a dictionary.

        This method processes each key-value pair in the given dictionary. If a value
        is a string, it strips leading and trailing whitespace from it. If a value is
        a dictionary, it applies the same stripping process recursively to that
        dictionary. The keys of the dictionary are also stripped of leading and trailing
        whitespace. Non-string values are included in the output without modification.

        Args:
            obj (dict): The dictionary whose keys and string values will have whitespace
                removed. It can be nested, with dictionaries as values, which will
                also be processed.

        Returns:
            dict:
                A new dictionary with all keys and string values stripped of leading
                and trailing whitespace. The structure of the dictionary is preserved.
        """
        return {
            key.strip(): GaussianErrorHandler._recursive_remove_space(value)
            if isinstance(value, dict)
            else value.strip()
            if isinstance(value, str)
            else value
            for key, value in obj.items()
        }

    @staticmethod
    def _update_route_params(route_params: dict, key: str, value: str | dict) -> dict:
        """
        Update Gaussian route parameters with new key-value pairs, handling nested
        structures.

        Args:
            route_params (dict): The dictionary of route parameters to be updated.
            key (str): The key in the route parameters to update or add.
            value (str | dict): The new value to set or add to the route parameters.
                This can be a string or a dictionary. If it is a dictionary, it is
                merged with the existing dictionary at `key`.

        Returns:
            dict:
                The updated route parameters.
        """
        obj = route_params.get(key, {})
        if not obj:
            route_params[key] = value
        elif isinstance(obj, str):
            update = {key: {obj: None, **value}} if isinstance(value, dict) else {key: {obj: None, value: None}}
            route_params.update(update)
        elif isinstance(obj, dict):
            route_params[key].update(value if isinstance(value, dict) else {value: None})
        return route_params

    @staticmethod
    def _int_keyword(route_params: dict[str, str | dict]) -> tuple[str, str | dict]:
        """
        Determine the keyword used for 'Integral' in the Gaussian route parameters of
        the input file. Possible keywords are 'int' and 'integral'. If neither keyword
        is found, an empty string is returned.

        Args:
            route_params (dict): The route parameters dictionary.

        Returns:
            tuple:
                The key ('int' or 'integral' or an empty string if neither is found),
                and the value associated with this key in `route_params`. If the key is
                not found, the second element in the tuple is an empty string.
        """
        if "int" in route_params:
            int_key = "int"
        elif "integral" in route_params:
            int_key = "integral"
        else:
            int_key = ""
        # int_key = 'int' if 'int' in route_params else 'integral'
        return int_key, route_params.get(int_key, "")

    @staticmethod
    def _int_grid(route_params: dict[str, str | dict]) -> bool:
        """
        Check if the integration grid used for numerical integrations matches specific
        options.

        Args:
            route_params (dict): The route parameters dictionary.

        Returns:
            bool:
                True if the integral grid parameter matches one of the predefined
                options, otherwise False.
        """
        _, int_value = GaussianErrorHandler._int_keyword(route_params)
        options = ["ultrafine", "ultrafinegrid", "99590"]

        if isinstance(int_value, str) and int_value in options:
            return True
        if isinstance(int_value, dict):
            if int_value.get("grid") in options:
                return True
            if set(int_value) & set(options):
                return True
        return False

    @staticmethod
    def convert_mem(mem: float, unit: str) -> float:
        """
        Convert memory size between different units to megabytes (MB).

        Args:
            mem (float): The memory size to convert.
            unit (str): The unit of the input memory size. Supported units include
                'kb', 'mb', 'gb', 'tb', and word units ('kw', 'mw', 'gw', 'tw'), or an
                empty string for default conversion (from words).

        Returns:
            float:
                The memory size in MB.
        """
        conversion = {
            "kb": 1 / 1000,
            "mb": 1,
            "gb": 1000,
            "tb": 1000**2,
            "": 7.63e-6,
            "kw": 7.63e-3,
            "mw": 7.63,
            "gw": 7.63e3,
            "tw": 7.63e6,
        }
        return mem * conversion[unit]

    @staticmethod
    def _find_dynamic_memory_allocated(link0_params: dict[str, str]) -> tuple[str | None, float | None]:
        """
        Find and convert the memory allocation from Gaussian link0 parameters. This
        method searches for the '%mem' key in the link0 parameters of a Gaussian job
        to determine the memory allocated for the job. It extracts the memory value
        and its unit, then converts the  memory allocation to MB. The default memory
        unit used in Gaussian is words, and this method accounts for different units
        specified in the memory string.

        Args:
            link0_params (dict): A dictionary of link0 parameters from a Gaussian input
                file.

        Returns:
            tuple:
                The memory key (None if '%mem' is not found) and the converted memory
                allocation in MB. If '%mem' is not found, the second element will be None.
        """
        mem_key = None
        dynamic_mem = None
        for k in link0_params:
            if k.lower() == "%mem":
                mem_key = k
                break
        if mem_key:
            dynamic_mem_str = link0_params[mem_key]
            # default memory unit in Gaussian is words
            dynamic_mem_str = dynamic_mem_str.lower()
            mem_unit = ""
            for unit in GaussianErrorHandler.MEM_UNITS:
                if unit in dynamic_mem_str:
                    mem_unit = unit
                    break
            dynamic_mem = float(dynamic_mem_str.strip(mem_unit))
            dynamic_mem = GaussianErrorHandler.convert_mem(dynamic_mem, mem_unit)
        return mem_key, dynamic_mem

    def _add_int(self) -> bool:
        """
        Check and update the integration grid setting ('int') in the Gaussian input
        file's route parameters to 'ultrafine', if necessary.

        Returns:
            bool: True if changes were made to the integration grid setting, False otherwise.
        """
        if not GaussianErrorHandler._int_grid(self.gin.route_parameters):
            # nothing int is set or is set to different values
            warning_msg = (
                "Changing the numerical integration grid. "
                "This will bring changes in the predicted "
                "total energy. It is necessary to use the same "
                "integration grid in all the calculations in "
                "the same study in order for the computed "
                "energies and molecular properties to be "
                "comparable."
            )

            int_key, int_value = GaussianErrorHandler._int_keyword(self.gin.route_parameters)
            if not int_value and GaussianErrorHandler._not_g16(self.gout):
                # if int keyword is missing and Gaussian version is 03 or
                # 09, set integration grid to ultrafine
                int_key = int_key or "int"
                self.logger.warning(warning_msg)
                self.gin.route_parameters[int_key] = "ultrafine"
                return True
            if isinstance(int_value, dict):
                # if int grid is set and is different from ultrafine,
                # set it to ultrafine (works when others int options are
                # specified)
                flag = False
                if "grid" in self.gin.route_parameters[int_key]:
                    flag = True
                for key in self.gin.route_parameters[int_key]:
                    if key in self.GRID_NAMES or self.grid_patt.match(key):
                        self.gin.route_parameters[int_key].pop(key)
                        flag = True
                        break
                if flag or GaussianErrorHandler._not_g16(self.gout):
                    self.logger.warning(warning_msg)
                    self.gin.route_parameters[int_key]["grid"] = "ultrafine"
                    return True
            if isinstance(int_value, str) and (int_value in self.GRID_NAMES or self.grid_patt.match(int_value)):
                # if int grid is set and is different from ultrafine,
                # set it to ultrafine (works when no other int options
                # are specified)
                self.logger.warning(warning_msg)
                self.gin.route_parameters[int_key] = "ultrafine"
                return True
            if GaussianErrorHandler._not_g16(self.gout):
                # if int grid is not specified, and Gaussian version is
                # not 16, update with ultrafine integral grid
                self.logger.warning(warning_msg)
                GaussianErrorHandler._update_route_params(self.gin.route_parameters, int_key, {"grid": "ultrafine"})
                return True
            return False
        return False

    @staticmethod
    def _not_g16(gout: GaussianOutput) -> bool:
        """
        Determine if the Gaussian version is not 16.

        Args:
        gout (GaussianOutput): A GaussianOutput object.

        Returns:
            bool: True if the Gaussian version is not 16, False otherwise.
        """
        return "16" not in gout.version  # type:ignore[attr-defined]

    @staticmethod
    def _monitor_convergence(data: dict[str, dict[str, Any]], directory: str = "./") -> None:
        """
        Plot and save a convergence graph for an optimization job as a function of the
        number of iterations.

        Parameters:
        data (dict): A dictionary containing two keys: 'values' and 'thresh'. 'values'
            is a dictionary where each key-value pair represents a parameter and its
            values across iterations. 'thresh' is a dictionary where each key-value pair
            represents a parameter and its threshold value. The convergence parameters
            are: 'max_force', 'rms_force', 'max_disp', and 'rms_disp'.
        directory (str, optional): The directory where the convergence plot image will
            be saved. Defaults to "./".
        """
        import matplotlib.pyplot as plt
        from matplotlib.ticker import MaxNLocator

        _fig, ax = plt.subplots(ncols=2, nrows=2, figsize=(12, 10))
        for i, (k, v) in enumerate(data["values"].items()):
            row = int(np.floor(i / 2))
            col = i % 2
            iters = range(len(v))
            ax[row, col].plot(iters, v, color="#cf3759", linewidth=2)
            ax[row, col].axhline(y=data["thresh"][k], linewidth=2, color="black", linestyle="--")
            ax[row, col].tick_params(which="major", length=8)
            ax[row, col].tick_params(axis="both", which="both", direction="in", labelsize=16)
            ax[row, col].set_xlabel("Iteration", fontsize=16)
            ax[row, col].set_ylabel(f"{k}", fontsize=16)
            ax[row, col].xaxis.set_major_locator(MaxNLocator(integer=True))
            ax[row, col].grid(ls="--", zorder=1)
        plt.tight_layout()
        plt.savefig(os.path.join(directory, "convergence.png"))

    def check(self, directory: str = "./") -> bool:
        """Check for errors in the Gaussian output file."""
        # TODO: this backups the original file instead of the actual one
        if "linear_bend" in self.errors:
            os.rename(
                os.path.join(directory, self.input_file + ".prev"),
                os.path.join(directory, self.input_file),
            )

        self.gin = GaussianInput.from_file(os.path.join(directory, self.input_file))
        self.gin.route_parameters = GaussianErrorHandler._recursive_lowercase(self.gin.route_parameters)
        assert isinstance(self.gin.route_parameters, dict)
        self.gin.route_parameters = GaussianErrorHandler._recursive_remove_space(self.gin.route_parameters)
        self.gout = GaussianOutput(os.path.join(directory, self.output_file))
        self.errors = set()
        error_patts = set()
        # TODO: move this to pymatgen?
        self.conv_data = {"values": {}, "thresh": {}}
        with zopen(os.path.join(directory, self.output_file), "rt") as f:
            for line in f:
                error_match = GaussianErrorHandler.error_patt.search(line)  # type:ignore[arg-type]
                mem_match = GaussianErrorHandler.recom_mem_patt.search(line)  # type:ignore[arg-type]
                if error_match:
                    patt = error_match.group(0)
                    error_patts.add(patt)
                    for pattern, error_key in GaussianErrorHandler.error_defs.items():
                        if re.match(pattern, patt):
                            self.errors.add(error_key)
                            break
                    # self.errors.add(GaussianErrorHandler.error_defs[patt])
                if mem_match:
                    mem = mem_match.group(1)
                    self.recom_mem = GaussianErrorHandler.convert_mem(float(mem), "mw")

                if self.check_convergence and "opt" in self.gin.route_parameters:
                    for k, v in GaussianErrorHandler.conv_criteria.items():
                        m = v.search(line)
                        if m:
                            if k not in self.conv_data["values"]:
                                self.conv_data["values"][k] = [m.group(2)]
                                self.conv_data["thresh"][k] = float(m.group(3))
                            else:
                                self.conv_data["values"][k].append(m.group(2))

        # TODO: it only plots after the job finishes, modify?
        if self.conv_data["values"] and all(len(v) >= 2 for v in self.conv_data["values"].values()):
            for k, v in self.conv_data["values"].items():
                # convert strings to float taking into account the
                # possibility of having ******** values
                self.conv_data["values"][k] = np.genfromtxt(np.array(v))
            GaussianErrorHandler._monitor_convergence(self.conv_data)
        for patt in error_patts:
            self.logger.error(patt)
        return len(self.errors) > 0

    def correct(self, directory: str = "./"):
        """Perform necessary actions to correct the errors in the Gaussian output."""
        actions: list[Any] = []
        # to avoid situations like 'linear_bend', where if we backup input_file,
        # it will not be the actual input used in the current calc
        # shutil.copy(self.input_file, f'{self.input_file}.backup')
        # backup_files = [self.input_file, self.output_file,
        #                 self.stderr_file]
        # checkpoint = glob.glob('*.[Cc][Hh][Kk]')
        # form_checkpoint = glob.glob('*.[Ff][Cc][Hh][Kk]')
        # png = glob.glob('convergence.png')
        # [backup_files.append(i[0]) for i in [checkpoint, form_checkpoint, png]
        #  if i]
        # backup(backup_files, self.prefix)
        # os.remove(f'{self.input_file}.backup')
        backup_files = [self.input_file, self.output_file, self.stderr_file, *BACKUP_FILES.values()]
        backup(backup_files, prefix=self.prefix, directory=directory)
        if "scf_convergence" in self.errors:
            self.gin.route_parameters = GaussianErrorHandler._update_route_params(self.gin.route_parameters, "scf", {})
            # if the SCF procedure has failed to converge
            if self.gin.route_parameters.get("scf", {}).get("maxcycle") != str(self.scf_max_cycles):
                # increase number of cycles if not already set or is different
                # from scf_max_cycles
                self.gin.route_parameters["scf"]["maxcycle"] = self.scf_max_cycles
                actions.append({"scf_max_cycles": self.scf_max_cycles})

            elif not {"xqc", "yqc", "qc"}.intersection(self.gin.route_parameters.get("scf", set())):
                # use an alternate SCF converger
                self.gin.route_parameters["scf"]["xqc"] = None
                actions.append({"scf_algorithm": "xqc"})

            elif self.job_type == "better_guess" and not GaussianErrorHandler.activate_better_guess:
                # try to get a better initial guess at a lower level of theory
                self.logger.info(
                    "SCF calculation failed. Switching to a lower "
                    "level of theory to get a better initial "
                    "guess of molecular orbitals"
                )
                # TODO: what if inputs don't work with scf_lot? e.g. extra_basis
                self.gin.functional = self.lower_functional
                self.gin.basis_set = self.lower_basis_set
                GaussianErrorHandler.activate_better_guess = True
                actions.append({"scf_level_of_theory": "better_scf_guess"})

            else:
                if self.job_type != "better_guess":
                    self.logger.info(
                        "Try to switch to better_guess job type to "
                        "generate a different initial guess using a "
                        "lower level of theory"
                    )
                else:
                    self.logger.info("SCF calculation failed. Exiting...")
                return {"errors": list(self.errors), "actions": None}

        elif "opt_steps" in self.errors:
            # int_actions = self._add_int()
            if self.gin.route_parameters.get("opt").get("maxcycles") != str(self.opt_max_cycles):
                self.gin.route_parameters["opt"]["maxcycles"] = self.opt_max_cycles
                if len(self.gout.structures) > 1:
                    self.gin._mol = self.gout.final_structure
                    actions.append({"structure": "from_final_structure"})
                actions.append({"opt_max_cycles": self.opt_max_cycles})

            elif self.check_convergence and all(v[-1] < v[0] for v in self.conv_data["values"].values()):
                self.gin._mol = self.gout.final_structure
                actions.append({"structure": "from_final_structure"})

            elif self._add_int():
                actions.append({"integral": "ultra_fine"})

            # elif int_actions:
            #     actions.append(int_actions)
            # TODO: check if the defined methods are clean
            # TODO: don't enter this if condition if g16 and ...

            elif self.job_type == "better_guess" and not GaussianErrorHandler.activate_better_guess:
                # TODO: check if the logic is correct since this is used with scf
                # try to get a better initial guess at a lower level of theory
                self.logger.info(
                    "Geometry optimization failed. Switching to a "
                    "lower level of theory to get a better "
                    "initial guess of molecular geometry"
                )
                self.gin.functional = self.lower_functional
                self.gin.basis_set = self.lower_basis_set
                GaussianErrorHandler.activate_better_guess = True
                actions.append({"opt_level_of_theory": "better_geom_guess"})

            else:
                if self.job_type != "better_guess":
                    self.logger.info(
                        "Try to switch to better_guess job type to "
                        "generate a different initial guess using a "
                        "lower level of theory"
                    )
                else:
                    self.logger.info("Geometry optimization failed. Exiting...")
                return {"errors": list(self.errors), "actions": None}

        elif "linear_bend" in self.errors:
            # if there is some linear bend around an angle in the geometry,
            # restart the job at the point it stopped while forcing Gaussian
            # to rebuild the set of redundant internals
            if not list(filter(re.compile(r"%[Cc][Hh][Kk]").match, self.gin.link0_parameters.keys())):
                raise KeyError("This remedy reads coords from checkpoint file. Consider adding CHK to link0_parameters")
            self.gin = GaussianInput(
                mol=None,
                charge=self.gin.charge,
                spin_multiplicity=self.gin.spin_multiplicity,
                title=self.gin.title,
                functional=self.gin.functional,
                basis_set=self.gin.basis_set,
                route_parameters=self.gin.route_parameters,
                input_parameters=self.gin.input_parameters,
                link0_parameters=self.gin.link0_parameters,
                dieze_tag=self.gin.dieze_tag,
                gen_basis=self.gin.gen_basis,
            )
            self.gin.route_parameters.update({"geom": "(checkpoint, newdefinition)"})
            actions.append({"coords": "rebuild_redundant_internals"})

        elif "solute_solvent_surface" in self.errors:
            # if non-convergence in the iteration of the PCM matrix is
            # encountered, change the type of molecular surface representing
            # the solute-solvent boundary
            # TODO: test
            input_parms = {
                key.lower() if isinstance(key, str) else key: value for key, value in self.gin.input_parameters.items()
            }
            if input_parms.get("surface", "none").lower() != "sas":
                GaussianErrorHandler._update_route_params(self.gin.route_parameters, "scrf", "read")
                self.gin.input_parameters.update({"surface": "SAS"})
                actions.append({"surface": "SAS"})
            else:
                self.logger.info("Not sure how to fix solute_solvent_surface_error if surface is already SAS!")
                return {"errors": [self.errors], "actions": None}

        elif "internal_coords" in self.errors:
            # check if optimization is requested to be performed in cartesian
            # coords. if not, set it while overwriting other possibly requested
            # coord systems, disable symmetry if applicable, and rerun
            # however, this will come at a higher computational cost
            if "opt" in self.gin.route_parameters and not any(
                n in (self.gin.route_parameters.get("opt") or {}) for n in ["cart", "cartesian"]
            ):
                GaussianErrorHandler._update_route_params(self.gin.route_parameters, "opt", "cartesian")
                if isinstance(self.gin.route_parameters["opt"], dict):
                    [self.gin.route_parameters["opt"].pop(i, None) for i in ["redundant", "zmatrix", "z-matrix"]]

                if (
                    not self.gin.route_parameters.get("nosymmetry")
                    or self.gin.route_parameters.get("symmetry") != "none"
                ):
                    self.gin.route_parameters["symmetry"] = "none"
                    actions.append({"symmetry": False})
                actions.append({"opt_cart_coords": True})
            else:
                self.logger.info(
                    "An error occurred in internal coordinates. "
                    "Your molecule might have 3 or more atoms "
                    "that are nearly linear making it difficult "
                    "to generate internal coordinates. Try to "
                    "modify your structure input?"
                )
                return {"errors": [self.errors], "actions": None}

        elif "zmatrix" in self.errors:
            with open(os.path.join(directory, self.input_file)) as gfile:
                lines = gfile.readlines()
                last_lines = lines[-2:]
            if set(last_lines) != {"\n"}:
                # if the required blank lines at the end of the input file are
                # missing, just rewrite the file
                self.logger.info("Missing blank line at the end of the input file.")
                actions.append({"blank_lines": "rewrite_input_file"})
            else:
                self.logger.info("Not sure how to fix zmatrix error. Check manually?")
                return {"errors": [self.errors], "actions": None}

        elif "coords" in self.errors:
            if "connectivity" in self.gin.route_parameters.get("geom"):
                self.logger.info("Explicit atom bonding is requested but no such input is provided")
                if isinstance(self.gin.route_parameters["geom"], dict) and len(self.gin.route_parameters["geom"]) > 1:
                    self.gin.route_parameters["geom"].pop("connectivity", None)
                else:
                    del self.gin.route_parameters["geom"]
                actions.append({"coords": "remove_connectivity"})
            else:
                self.logger.info("Missing connectivity info. Not sure how to fix this error. Exiting!")
                return {"errors": [self.errors], "actions": None}

        elif "found_coords" in self.errors:
            if self.gin.molecule and any(
                key in self.gin.route_parameters.get("geom", {}) for key in ["checkpoint", "check", "allcheck"]
            ):
                # if coords are found in the input and the user chooses to read
                # the molecule specification from the checkpoint file,
                # remove mol
                self.gin._mol = None
                actions.append({"mol": "remove_from_input"})
            else:
                self.logger.info("Not sure why atom specifications should not be found in the input. Examine manually!")
                return {"errors": [self.errors], "actions": None}

        elif "coord_inputs" in self.errors:
            if (
                any(key in self.gin.route_parameters.get("opt", {}) for key in ["z-matrix", "zmatrix"])
                and self.cart_coords
            ):
                # if molecule is specified in xyz format, but the user chooses
                # to perform the optimization using internal coordinates,
                # switch to z-matrix format
                self.cart_coords = False
                actions.append({"coords": "use_zmatrix_format"})
            else:
                # error cannot be fixed automatically. Return None for actions
                self.logger.info(
                    "Not sure how to fix problem with z-matrix "
                    "optimization if coords are already input in"
                    "z-matrix format. Examine manually!"
                )
                return {"errors": [self.errors], "actions": None}

        elif "missing_mol" in self.errors:
            if (
                not self.gin.molecule
                and "read" in self.gin.route_parameters.get("guess")
                and not any(
                    key in self.gin.route_parameters.get("geom", {}) for key in ["checkpoint", "check", "allcheck"]
                )
            ):
                # if molecule is not specified and the user requests that the
                # initial guess be read from the checkpoint file but forgot to
                # take the geom from the checkpoint file, add geom=check
                if not glob.glob("*.[Cc][Hh][Kk]"):
                    raise FileNotFoundError("This remedy reads geometry from checkpoint file. This file is missing!")
                GaussianErrorHandler._update_route_params(self.gin.route_parameters, "geom", "check")
                self.gin.route_parameters["geom"] = "check"
                actions.append({"mol": "get_from_checkpoint"})
            else:
                # error cannot be fixed automatically. Return None for actions
                self.logger.info("Molecule is not found in the input file. Fix manually!")
                # TODO: check if logger.info is enough here or return is needed
                return {"errors": list(self.errors), "actions": None}

        elif any(err in self.errors for err in ["empty_file", "bad_file"]):
            self.logger.error("Required checkpoint file is bad. Fix manually!")
            return {"errors": list(self.errors), "actions": None}

        elif "missing_file" in self.errors:
            self.logger.error("Could not find the required file. Fix manually!")
            return {"errors": list(self.errors), "actions": None}

        elif "syntax" in self.errors:
            # error cannot be fixed automatically. Return None for actions
            self.logger.info("A syntax error was detected in the input file. Fix manually!")
            return {"errors": list(self.errors), "actions": None}

        elif "insufficient_mem" in self.errors:
            mem_key, dynamic_mem = GaussianErrorHandler._find_dynamic_memory_allocated(self.gin.link0_parameters)
            if dynamic_mem and self.recom_mem and dynamic_mem < self.recom_mem:
                # this assumes that 1.5*minimum required memory is available
                mem = math.ceil(self.recom_mem * 1.5)
                self.gin.link0_parameters[mem_key] = f"{mem}MB"
                actions.append({"memory": "increase_to_gaussian_recommendation"})
            else:
                self.logger.info("Check job memory requirements manually and set as needed.")
                return {"errors": list(self.errors), "actions": None}

        else:
            self.logger.info("Must have gotten an error that is parsed but not handled yet. Fix manually!")
            return {"errors": list(self.errors), "actions": None}

        os.rename(
            os.path.join(directory, self.input_file),
            os.path.join(directory, self.input_file + ".prev"),
        )
        self.gin.write_file(os.path.join(directory, self.input_file), self.cart_coords)
        # TODO: ADDED
        if os.path.exists(os.path.join(directory, self.input_file) + ".wt"):
            shutil.copyfile(
                os.path.join(directory, self.input_file),
                os.path.join(directory, self.input_file + ".wt"),
            )
        return {"errors": list(self.errors), "actions": actions}


class WallTimeErrorHandler(ErrorHandler):
    """
    Check if a run is nearing the walltime. If so, terminate the job and restart from
    the last .rwf file. A job is considered to be nearing the walltime if the remaining
    time is less than or equal to the buffer time.
    """

    is_monitor: bool = True

    def __init__(
        self,
        wall_time: int,
        buffer_time: int,
        input_file: str,
        output_file: str,
        stderr_file: str = "stderr.txt",
        prefix: str = "error",
    ):
        """
        Initialize the WalTimeErrorHandler class.

        Args:
            wall_time (int): The total wall time for the job in seconds.
            buffer_time (int): The buffer time in seconds. If the remaining time is less
                than or equal to the buffer time, the job is considered to be nearing the
                walltime and will be terminated.
            input_file (str): The name of the input file for the Gaussian job.
            output_file (str): The name of the output file for the Gaussian job.
            stderr_file (str): The name of the standard error file for the Gaussian job.
                Defaults to 'stderr.txt'.
            prefix (str): The prefix to use for the backup files. Defaults to error,
                which means a series of error.1.tar.gz, error.2.tar.gz, ... will be
                generated.
        """
        self.wall_time = wall_time
        self.buffer_time = buffer_time
        self.input_file = input_file
        self.output_file = output_file
        self.stderr_file = stderr_file
        self.prefix = prefix
        self.logger: logging.Logger = logging.getLogger(self.__class__.__name__)
        logging.basicConfig(level=logging.INFO)

        now_ = datetime.datetime.now()
        now_str = datetime.datetime.strftime(now_, "%a %b %d %H:%M:%S UTC %Y")
        init_time_str = os.environ.get("JOB_START_TIME", now_str)
        os.environ["JOB_START_TIME"] = init_time_str
        self.init_time = datetime.datetime.strptime(init_time_str, "%a %b %d %H:%M:%S %Z %Y")

    def check(self, directory: str = "./") -> bool:
        """Check if the job is nearing the walltime. If so, return True, else False."""
        if self.wall_time:
            run_time = datetime.datetime.now() - self.init_time
            remaining_time = self.wall_time - run_time.total_seconds()
            if remaining_time <= self.buffer_time:
                return True
        return False

    def correct(self, directory: str = "./") -> dict:
        """Perform the corrections."""
        # TODO: when using restart, the rwf file might be in a different dir
        backup_files = [self.input_file, self.output_file, self.stderr_file, *BACKUP_FILES.values()]
        backup(backup_files, prefix=self.prefix, directory=directory)
        if glob.glob(os.path.join(directory, BACKUP_FILES["rwf"])):
            rwf = glob.glob(os.path.join(directory, BACKUP_FILES["rwf"]))[0]
            gin = GaussianInput.from_file(os.path.join(directory, self.input_file))
            # TODO: check if rwf is already there like RWF or Rwf or ...
            # gin.link0_parameters.update({'%rwf': rwf})
            # gin.route_parameters = {'Restart': None}
            # os.rename(self.input_file, self.input_file + '.prev')
            input_str = [f"%rwf={rwf}"] + [f"{i}={j}" for i, j in gin.link0_parameters.items()]
            input_str.append(f"{gin.dieze_tag} Restart\n\n")
            with open(os.path.join(directory, self.input_file + ".wt"), "w") as f:
                f.write("\n".join(input_str))
            return {"errors": ["wall_time_limit"], "actions": None}
        self.logger.info(
            "Wall time handler requires a read-write gaussian file to be available. No such file is found."
        )
        return {"errors": ["Walltime reached but not rwf file found"], "actions": None}



"""Implements various validators, e.g., check if vasprun.xml is valid, for VASP."""

from __future__ import annotations

import logging
import os
from collections import deque

from pymatgen.io.vasp import Chgcar, Incar

from custodian.custodian import Validator
from custodian.vasp.io import load_outcar, load_vasprun


class VasprunXMLValidator(Validator):
    """Checks that a valid vasprun.xml was generated."""

    def __init__(self, output_file: str = "vasp.out", stderr_file: str = "std_err.txt") -> None:
        """
        Args:
            output_file (str): Name of file VASP standard output is directed to.
                Defaults to "vasp.out".
            stderr_file (str): Name of file VASP standard error is direct to.
                Defaults to "std_err.txt".
        """
        self.output_file = output_file
        self.stderr_file = stderr_file
        self.logger = logging.getLogger(type(self).__name__)

    def check(self, directory="./") -> bool:
        """Check for errors."""
        try:
            load_vasprun(os.path.join(directory, "vasprun.xml"))
        except Exception:
            exception_context: dict[str, str | float] = {}

            if os.path.isfile(os.path.join(directory, self.output_file)):
                with open(os.path.join(directory, self.output_file)) as output_file:
                    output_file_tail = deque(output_file, maxlen=10)
                exception_context["output_file_tail"] = "".join(output_file_tail)

            if os.path.isfile(os.path.join(directory, self.stderr_file)):
                with open(os.path.join(directory, self.stderr_file)) as stderr_file:
                    stderr_file_tail = deque(stderr_file, maxlen=10)
                exception_context["stderr_file_tail"] = "".join(stderr_file_tail)

            if os.path.isfile(os.path.join(directory, "vasprun.xml")):
                stat = os.stat(os.path.join(directory, "vasprun.xml"))
                exception_context["vasprun_st_size"] = stat.st_size
                exception_context["vasprun_st_atime"] = stat.st_atime
                exception_context["vasprun_st_mtime"] = stat.st_mtime
                exception_context["vasprun_st_ctime"] = stat.st_ctime

                with open(os.path.join(directory, "vasprun.xml")) as vasprun:
                    vasprun_tail = deque(vasprun, maxlen=10)
                exception_context["vasprun_tail"] = "".join(vasprun_tail)

            self.logger.exception("Failed to load vasprun.xml", extra=exception_context)

            return True
        return False


class VaspFilesValidator(Validator):
    """
    Check for existence of some of the files that VASP
        normally create upon running.
    """

    def __init__(self) -> None:
        """Dummy init."""

    def check(self, directory="./"):
        """Check for error."""
        return any(not os.path.isfile(os.path.join(directory, vfile)) for vfile in ("CONTCAR", "OSZICAR", "OUTCAR"))


class VaspNpTMDValidator(Validator):
    """
    Check NpT-AIMD settings is loaded by VASP compiled with -Dtbdyn.
    Currently, VASP only have Langevin thermostat (MDALGO = 3) for NpT ensemble.
    """

    def __init__(self) -> None:
        """Dummy init."""

    def check(self, directory="./"):
        """Check for error."""
        incar = Incar.from_file(os.path.join(directory, "INCAR"))
        is_npt = incar.get("MDALGO") == 3
        if not is_npt:
            return False

        outcar = load_outcar(os.path.join(directory, "OUTCAR"))
        patterns = {"MDALGO": r"MDALGO\s+=\s+([\d]+)"}
        outcar.read_pattern(patterns=patterns)
        return outcar.data["MDALGO"] != [["3"]]


class VaspAECCARValidator(Validator):
    """Check if the data in the AECCAR is corrupted."""

    def __init__(self) -> None:
        """Dummy init."""

    def check(self, directory="./"):
        """Check for error."""
        aeccar0 = Chgcar.from_file(os.path.join(directory, "AECCAR0"))
        aeccar2 = Chgcar.from_file(os.path.join(directory, "AECCAR2"))
        aeccar = aeccar0 + aeccar2
        return check_broken_chgcar(aeccar)


def check_broken_chgcar(chgcar, diff_thresh=None) -> bool:
    """
    Check if the charge density file is corrupt
    Args:
        chgcar (Chgcar): Chgcar-like object.
        diff_thresh (Float): Threshold for diagonal difference.
            None means we won't check for this.
    """
    chgcar_data = chgcar.data["total"]
    if (chgcar_data < 0).sum() > 100:
        # a decent bunch of the values are negative this for sure means a broken charge density
        return True

    if diff_thresh:
        """
        If any one diagonal difference accounts for more than a particular portion of
        the total difference between highest and lowest density.
        When we are looking at AECCAR data, since the charge density is so high near the core
        and we have a course grid, this threshold can be as high as 0.99
        """
        diff = chgcar_data[:-1, :-1, :-1] - chgcar_data[1:, 1:, 1:]
        if diff.max() / (chgcar_data.max() - chgcar_data.min()) > diff_thresh:
            return True

    return False



# Copyright (c) Materials Virtual Lab.
# Distributed under the terms of the BSD License.

"""A yaml based Custodian job runner. Allows for multi-step jobs with modifications along the way."""

import argparse
import logging
import sys

from monty.serialization import loadfn

from custodian.custodian import Custodian

example_yaml = """
# This is an example of a Custodian yaml spec file. It shows how you can specify
# a double relaxation followed by a static calculation. Minor modifications
# would allow very customizable calculations, though this is obviously not meant
# for highly complex workflows. For those, you will need to code and usage of
# FireWorks is highly recommended.


# Specifies a list of jobs to run.
# Each job is specified by a `jb: <full class path>` with parameters specified
# via the params dict.

jobs:
- jb: custodian.vasp.jobs.VaspJob
  params:
    final: False
    suffix: .relax1
- jb: custodian.vasp.jobs.VaspJob
  params:
    final: False
    suffix: .relax2
    settings_override:
    - {"file": "CONTCAR", "action": {"_file_copy": {"dest": "POSCAR"}}}
- jb: custodian.vasp.jobs.VaspJob
  params:
    final: True
    suffix: .static3
    settings_override:
    - {"file": "CONTCAR", "action": {"_file_copy": {"dest": "POSCAR"}}}
    - {"dict": "INCAR", "action": {"_set": {"NSW": 0}}}


# This key specifies parameters common to all jobs.
# Keys starting with $ are expanded to the environmental values.
# The example below means the parameter vasp_cmd is set to the value with
# $PBS_NODEFILE expanded.

jobs_common_params:
  $vasp_cmd: ["mpirun", "-machinefile", "$PBS_NODEFILE", "-np", "24", "vasp"]


# Specifies a list of error handlers in the same format as jobs.
handlers:
- hdlr: custodian.vasp.handlers.VaspErrorHandler
- hdlr: custodian.vasp.handlers.AliasingErrorHandler
- hdlr: custodian.vasp.handlers.MeshSymmetryErrorHandler


# Specifies a list of error handlers in the same format as jobs.
validators:
- vldr: custodian.vasp.validators.VasprunXMLValidator


# This sets all custodian running parameters.
custodian_params:
  max_errors: 10
  scratch_dir: /tmp
  gzipped_output: True
  checkpoint: True
"""


def run(args) -> None:
    """Perform a single run."""
    FORMAT = "%(asctime)s %(message)s"
    logging.basicConfig(format=FORMAT, level=logging.INFO, filename="run.log")
    logging.info(f"Spec file is {args.spec_file}")
    dct = loadfn(args.spec_file[0])
    c = Custodian.from_spec(dct)
    c.run()


def print_example(args) -> None:
    """Print the example_yaml."""
    print(example_yaml)


def main() -> None:
    """Main method."""
    parser = argparse.ArgumentParser(
        description="""
    cstdn is a convenient script to run custodian style jobs using a
    simple YAML spec.""",
        epilog="""Author: Shyue Ping Ong""",
    )

    subparsers = parser.add_subparsers()

    prun = subparsers.add_parser("run", help="Run custodian.")
    prun.add_argument("spec_file", metavar="spec_file", type=str, nargs=1, help="YAML/JSON spec file.")
    prun.set_defaults(func=run)

    prun = subparsers.add_parser(
        "example",
        help="Print examples. Right now, there is only one example for VASP double relaxation.",
    )
    prun.set_defaults(func=print_example)

    args = parser.parse_args()

    try:
        _ = args.func
    except AttributeError:
        parser.print_help()
        sys.exit(0)
    args.func(args)


if __name__ == "__main__":
    main()



"""This module implements error handlers for QChem runs."""

import os

from pymatgen.io.qchem.inputs import QCInput
from pymatgen.io.qchem.outputs import QCOutput

from custodian.custodian import ErrorHandler
from custodian.utils import backup

try:
    from openbabel import openbabel as ob
except ImportError:
    ob = None

__author__ = "Samuel Blau, Brandon Wood, Shyam Dwaraknath, Ryan Kingsbury"
__copyright__ = "Copyright 2018, The Materials Project"
__version__ = "0.1"
__maintainer__ = "Samuel Blau"
__email__ = "samblau1@gmail.com"
__status__ = "Alpha"
__date__ = "3/26/18"
__credits__ = "Xiaohui Qu"


class QChemErrorHandler(ErrorHandler):
    """
    Master QChemErrorHandler class that handles a number of common errors
    that occur during QChem runs.
    """

    is_monitor = False

    def __init__(
        self,
        input_file="mol.qin",
        output_file="mol.qout",
        scf_max_cycles=100,
        geom_max_cycles=200,
    ) -> None:
        """Initialize the error handler from a set of input and output files.

        Args:
            input_file (str): Name of the QChem input file.
            output_file (str): Name of the QChem output file.
            scf_max_cycles (int): The max iterations to set to fix SCF failure.
            geom_max_cycles (int): The max iterations to set to fix geometry
                optimization failure.
        """
        self.input_file = input_file
        self.output_file = output_file
        self.scf_max_cycles = scf_max_cycles
        self.geom_max_cycles = geom_max_cycles
        self.outdata = None
        self.errors: list[str] = []
        self.opt_error_history: list[str] = []

    def check(self, directory="./"):
        """Checks output file for errors."""
        self._output_path = os.path.join(directory, self.output_file)
        self.outdata = QCOutput(self._output_path).data
        self.errors = self.outdata.get("errors")
        self.warnings = self.outdata.get("warnings")
        # If we aren't out of optimization cycles, but we were in the past, reset the history
        if "out_of_opt_cycles" not in self.errors and len(self.opt_error_history) > 0:
            self.opt_error_history = []
        # If we're out of optimization cycles and we have unconnected fragments, no need to handle any errors
        if "out_of_opt_cycles" in self.errors and self.outdata["structure_change"] == "unconnected_fragments":
            return False
        return len(self.errors) > 0

    def correct(self, directory="./"):
        """Perform corrections."""
        self._input_path = os.path.join(directory, self.input_file)
        self._output_path = os.path.join(directory, self.output_file)
        backup({self._input_path, self._output_path})
        actions = []
        self.qcinp = QCInput.from_file(self._input_path)

        if "SCF_failed_to_converge" in self.errors:
            # Given defaults, the first handlers will typically be skipped.
            if self.qcinp.rem.get("s2thresh", "14") != "16":
                self.qcinp.rem["s2thresh"] = "16"
                actions.append({"s2thresh": "16"})
            if (
                int(self.qcinp.rem.get("max_scf_cycles", 50)) < self.scf_max_cycles
                or self.qcinp.rem.get("thresh", "10") != "14"
            ):
                if int(self.qcinp.rem.get("max_scf_cycles", 50)) < self.scf_max_cycles:
                    self.qcinp.rem["max_scf_cycles"] = self.scf_max_cycles
                    actions.append({"max_scf_cycles": self.scf_max_cycles})
                if self.qcinp.rem.get("thresh", "10") != "14":
                    self.qcinp.rem["thresh"] = "14"
                    actions.append({"thresh": "14"})
            # Turn on GDM
            elif self.qcinp.rem.get("scf_algorithm", "diis").lower() == "diis":
                self.qcinp.rem["scf_algorithm"] = "gdm"
                actions.append({"scf_algorithm": "gdm"})
                self.qcinp.rem["max_scf_cycles"] = "500"
                actions.append({"max_scf_cycles": "500"})
            # Ensure we have 500 cycles for GDM
            elif (
                self.qcinp.rem.get("scf_algorithm", "diis").lower() == "gdm"
                and self.qcinp.rem.get("max_scf_cycles") != "500"
            ):
                self.qcinp.rem["max_scf_cycles"] = "500"
                actions.append({"max_scf_cycles": "500"})
            # Try forcing a new initial guess at each iteration
            elif (
                self.qcinp.rem.get("scf_guess_always", "none").lower() != "true"
                and "molecule_from_last_geometry" in self.outdata
            ):
                self.qcinp.rem["scf_guess_always"] = "true"
                actions.append({"scf_guess_always": "true"})
            else:
                print("No remaining SCF error handlers!")

        elif "out_of_opt_cycles" in self.errors:
            # Given defaults, the first two handlers will typically be skipped.
            if self.qcinp.rem.get("s2thresh", "14") != "16":
                self.qcinp.rem["s2thresh"] = "16"
                actions.append({"s2thresh": "16"})
            if self.qcinp.rem.get("thresh", "10") != "14":
                self.qcinp.rem["thresh"] = "14"
                actions.append({"thresh": "14"})
            # Check number of opt cycles. If less than geom_max_cycles, increase to that value
            # set last geom as new starting geom, make sure using DIIS_GDM for SCF and rerun.
            elif int(self.qcinp.rem.get("geom_opt_max_cycles", 50)) < self.geom_max_cycles:
                self.qcinp.rem["geom_opt_max_cycles"] = self.geom_max_cycles
                if str(self.qcinp.rem.get("geom_opt2", "none")) == "3" or self.outdata["version"] == "6":
                    self.qcinp.geom_opt["maxiter"] = self.geom_max_cycles  # pylint: disable=unsupported-assignment-operation
                actions.append({"geom_max_cycles:": self.geom_max_cycles})
                if "molecule_from_last_geometry" in self.outdata:
                    self.qcinp.molecule = self.outdata.get("molecule_from_last_geometry")
                    actions.append({"molecule": "molecule_from_last_geometry"})

            # Often can just get convergence by restarting from the geometry of the last cycle.
            # But we'll also save any structural changes that happened along the way.
            else:
                self.opt_error_history += [self.outdata["structure_change"]]
                if len(self.opt_error_history) > 1 and self.opt_error_history[-1] == "no_change":
                    # If no structural changes occurred in two consecutive optimizations,
                    # and we still haven't converged, then just exit. This is most common
                    # if two species are flying away from each other.
                    return {
                        "errors": self.errors,
                        "actions": None,
                        "opt_error_history": self.opt_error_history,
                    }
                self.qcinp.molecule = self.outdata.get("molecule_from_last_geometry")
                actions.append({"molecule": "molecule_from_last_geometry"})
                # Using GDM for SCF convergence also often helps.
                if self.qcinp.rem.get("scf_algorithm", "diis").lower() == "diis":
                    self.qcinp.rem["scf_algorithm"] = "gdm"
                    actions.append({"scf_algorithm": "gdm"})
                    self.qcinp.rem["max_scf_cycles"] = "500"
                    actions.append({"max_scf_cycles": "500"})

        elif "unable_to_determine_lamda" in self.errors:
            # Given defaults, the first two handlers will typically be skipped.
            if self.qcinp.rem.get("s2thresh", "14") != "16":
                self.qcinp.rem["s2thresh"] = "16"
                actions.append({"s2thresh": "16"})
            if self.qcinp.rem.get("thresh", "10") != "14":
                self.qcinp.rem["thresh"] = "14"
                actions.append({"thresh": "14"})
            elif "molecule_from_last_geometry" in self.outdata:
                self.qcinp.molecule = self.outdata.get("molecule_from_last_geometry")
                actions.append({"molecule": "molecule_from_last_geometry"})
                if self.qcinp.rem.get("scf_algorithm", "diis").lower() == "diis":
                    self.qcinp.rem["scf_algorithm"] = "gdm"
                    actions.append({"scf_algorithm": "gdm"})
                    self.qcinp.rem["max_scf_cycles"] = "500"
                    actions.append({"max_scf_cycles": "500"})
            else:
                print("Not sure how to fix Lambda error in this case!")

        elif "back_transform_error" in self.errors or "svd_failed" in self.errors:
            # Given defaults, the first two handlers will typically be skipped.
            if self.qcinp.rem.get("s2thresh", "14") != "16":
                self.qcinp.rem["s2thresh"] = "16"
                actions.append({"s2thresh": "16"})
            if self.qcinp.rem.get("thresh", "10") != "14":
                self.qcinp.rem["thresh"] = "14"
                actions.append({"thresh": "14"})
            elif "molecule_from_last_geometry" in self.outdata:
                self.qcinp.molecule = self.outdata.get("molecule_from_last_geometry")
                actions.append({"molecule": "molecule_from_last_geometry"})

            elif str(self.qcinp.rem.get("geom_opt2", "none")) == "3":
                self.qcinp.rem.pop("geom_opt2", None)
                self.qcinp.geom_opt = None
                actions.append({"geom_opt2": "deleted"})

            elif self.outdata["version"] == "6" and self.qcinp.rem.get("geom_opt_driver", "libopt3") != "optimize":
                if self.qcinp.geom_opt["coordinates"] == "redundant":
                    self.qcinp.geom_opt["coordinates"] = "delocalized"  # pylint: disable=unsupported-assignment-operation
                    actions.append({"coordinates": "delocalized"})
                    if self.qcinp.geom_opt.get("initial_hessian", "none") != "read":
                        self.qcinp.geom_opt["initial_hessian"] = "model"
                        actions.append({"initial_hessian": "model"})
                elif self.qcinp.geom_opt["coordinates"] == "delocalized":
                    self.qcinp.geom_opt["coordinates"] = "cartesian"  # pylint: disable=unsupported-assignment-operation
                    actions.append({"coordinates": "cartesian"})
                    if self.qcinp.geom_opt.get("initial_hessian", "none") == "model":
                        del self.qcinp.geom_opt["initial_hessian"]
                        actions.append({"initial_hessian": "deleted"})

        elif "premature_end_FileMan_error" in self.errors:
            # Given defaults, the first two handlers will typically be skipped.
            if self.qcinp.rem.get("s2thresh", "14") != "16":
                self.qcinp.rem["s2thresh"] = "16"
                actions.append({"s2thresh": "16"})
            if self.qcinp.rem.get("thresh", "10") != "14":
                self.qcinp.rem["thresh"] = "14"
                actions.append({"thresh": "14"})
            elif self.qcinp.rem.get("job_type") == "opt" or self.qcinp.rem.get("job_type") == "optimization":
                if self.qcinp.rem.get("scf_guess_always", "none").lower() != "true":
                    self.qcinp.rem["scf_guess_always"] = "true"
                    actions.append({"scf_guess_always": "true"})
                else:
                    print("Don't know how to fix a FileMan error for an opt while always generating a new SCF guess!")
            elif self.qcinp.rem.get("job_type") == "freq" or self.qcinp.rem.get("job_type") == "frequency":
                self.qcinp.rem["cpscf_nseg"] = str(self.outdata["cpscf_nseg"] + 1)
                actions.append({"cpscf_nseg": str(self.outdata["cpscf_nseg"] + 1)})

        elif "hessian_eigenvalue_error" in self.errors:
            if self.qcinp.rem.get("s2thresh", "14") != "16":
                self.qcinp.rem["s2thresh"] = "16"
                actions.append({"s2thresh": "16"})
            if self.qcinp.rem.get("thresh", "10") != "14":
                self.qcinp.rem["thresh"] = "14"
                actions.append({"thresh": "14"})
            else:
                print("Not sure how to fix hessian_eigenvalue_error if thresh is already 14!")

        elif "NLebdevPts" in self.errors:
            # this error should only be possible if resp_charges or esp_charges is set
            if self.qcinp.rem.get("resp_charges") or self.qcinp.rem.get("esp_charges"):
                # This error is caused by insufficient no. of Lebedev points on
                # the grid used to compute RESP charges
                # Increase the density of points on the Lebedev grid using the
                # esp_surface_density argument (see manual >= v5.4)
                # the default value is 500 (=0.001 Angstrom)
                # or disable RESP charges as a last resort
                if int(self.qcinp.rem.get("esp_surface_density", 500)) >= 500:
                    self.qcinp.rem["esp_surface_density"] = "250"
                    actions.append({"esp_surface_density": "250"})
                elif int(self.qcinp.rem.get("esp_surface_density", 250)) >= 250:
                    self.qcinp.rem["esp_surface_density"] = "125"
                    actions.append({"esp_surface_density": "125"})
                elif int(self.qcinp.rem.get("esp_surface_density", 125)) >= 125:
                    # switch from Lebedev mode to spherical harmonics mode
                    if self.qcinp.rem.get("resp_charges"):
                        self.qcinp.rem["resp_charges"] = "2"
                        actions.append({"resp_charges": "2"})
                    if self.qcinp.rem.get("esp_charges"):
                        self.qcinp.rem["esp_charges"] = "2"
                        actions.append({"esp_charges": "2"})
                else:
                    if self.qcinp.rem.get("resp_charges"):
                        self.qcinp.rem["resp_charges"] = "false"
                        actions.append({"resp_charges": "false"})
                    if self.qcinp.rem.get("esp_charges"):
                        self.qcinp.rem["esp_charges"] = "false"
                        actions.append({"esp_charges": "false"})
            else:
                raise RuntimeError("Not sure how to fix NLebdevPts error if resp_charges is disabled!")

        elif "failed_to_transform_coords" in self.errors:
            if self.qcinp.rem.get("thresh", "10") != "14":
                self.qcinp.rem["thresh"] = "14"
                actions.append({"thresh": "14"})
            if self.qcinp.rem.get("s2thresh", "14") != "16":
                self.qcinp.rem["s2thresh"] = "16"
                actions.append({"s2thresh": "16"})
            # Check for symmetry flag in rem. If not False, set to False and rerun.
            if not self.qcinp.rem.get("sym_ignore") or self.qcinp.rem.get("symmetry"):
                self.qcinp.rem["sym_ignore"] = "true"
                self.qcinp.rem["symmetry"] = "false"
                actions += ({"sym_ignore": "true"}, {"symmetry": "false"})
            else:
                print("Not sure how else to fix a failed coordinate transformation")

        elif "failed_cpscf" in self.errors:
            # For large systems, cpscf errors can often be resolved by forcing QChem to break up
            # the solving into more sections
            self.qcinp.rem["cpscf_nseg"] = str(self.outdata["cpscf_nseg"] + 1)
            actions.append({"cpscf_nseg": str(self.outdata["cpscf_nseg"] + 1)})

        elif "bad_old_nbo6_rem" in self.errors:
            # "run_nbo6" has to change to "nbo_external" in QChem 5.4.2 and later
            del self.qcinp.rem["run_nbo6"]
            self.qcinp.rem["nbo_external"] = "true"
            actions += ({"run_nbo6": "deleted"}, {"nbo_external": "true"})

        elif "bad_new_nbo_external_rem" in self.errors:
            # Have to use "run_nbo6" instead of "nbo_external" for QChem 5.4.1 or earlier
            del self.qcinp.rem["nbo_external"]
            self.qcinp.rem["run_nbo6"] = "true"
            actions += ({"nbo_external": "deleted"}, {"run_nbo6": "true"})

        elif "esp_chg_fit_error" in self.errors:
            # this error should only be possible if resp_charges or esp_charges is set
            if self.qcinp.rem.get("resp_charges") or self.qcinp.rem.get("esp_charges"):
                if self.qcinp.rem.get("resp_charges"):
                    self.qcinp.rem["resp_charges"] = "false"
                    actions.append({"resp_charges": "false"})
                if self.qcinp.rem.get("esp_charges"):
                    self.qcinp.rem["esp_charges"] = "false"
                    actions.append({"esp_charges": "false"})
            else:
                print("Not sure how to fix ESPChgFit error if resp_charges is disabled!")

        elif "probably_out_of_memory" in self.errors:
            # A frequency job that probably needs to be split into more cpscf segments
            # but did not exit gracefully
            if self.outdata["cpscf_nseg"] == 1:
                self.qcinp.rem["cpscf_nseg"] = "3"
                actions.append({"cpscf_nseg": "3"})
            else:
                self.qcinp.rem["cpscf_nseg"] = str(self.outdata["cpscf_nseg"] + 1)
                actions.append({"cpscf_nseg": str(self.outdata["cpscf_nseg"] + 1)})

        elif "gdm_neg_precon_error" in self.errors:
            if "molecule_from_last_geometry" in self.outdata:
                self.qcinp.molecule = self.outdata.get("molecule_from_last_geometry")
                actions.append({"molecule": "molecule_from_last_geometry"})
            else:
                print("Not sure how to fix gdm_neg_precon_error on the first SCF!")

        elif "mem_static_too_small" in self.errors:
            # mem_static should never exceed 2000 MB according to the Q-Chem manual
            self.qcinp.rem["mem_static"] = "2000"
            actions.append({"mem_static": "2000"})

        elif "mem_total_too_small" in self.errors:
            print(f"Run on a node with more memory! Current mem_total = {self.outdata['mem_total']}")
            return {"errors": self.errors, "actions": None}

        elif "basis_not_supported" in self.errors:
            print("Specify a different basis set. At least one of the atoms is not supported.")
            return {"errors": self.errors, "actions": None}

        elif "input_file_error" in self.errors:
            print("Something is wrong with the input file. Examine error message by hand.")
            return {"errors": self.errors, "actions": None}

        elif any(
            err in self.errors
            for err in ("failed_to_read_input", "read_molecule_error", "never_called_qchem", "licensing_error")
        ):
            # Almost certainly just a temporary problem that will not be encountered again. Rerun job as-is.
            actions.append({"rerun_job_no_changes": True})

        elif "unknown_error" in self.errors:
            if self.qcinp.rem.get("scf_guess", "none").lower() == "read":
                del self.qcinp.rem["scf_guess"]
                actions.append({"scf_guess": "deleted"})
            elif self.qcinp.rem.get("thresh", "10") != "14":
                self.qcinp.rem["thresh"] = "14"
                actions.append({"thresh": "14"})
            else:
                print("Unknown error. Examine output and log files by hand.")
                return {"errors": self.errors, "actions": None}

        else:
            # You should never get here. If correct is being called then errors should have at least one entry,
            # in which case it should have been caught by the if/elifs above.
            print("Errors:", self.errors)
            print("Must have gotten an error which is correctly parsed but not included in the handler. FIX!!!")
            return {"errors": self.errors, "actions": None}

        if {"molecule": "molecule_from_last_geometry"} in actions and str(
            self.qcinp.rem.get("geom_opt_hessian")
        ).lower() == "read":
            del self.qcinp.rem["geom_opt_hessian"]
            actions.append({"geom_opt_hessian": "deleted"})
        if (
            {"molecule": "molecule_from_last_geometry"} in actions
            and self.outdata["version"] == "6"
            and "initial_hessian" in self.qcinp.geom_opt
        ) and str(self.qcinp.geom_opt["initial_hessian"]).lower() == "read":
            del self.qcinp.geom_opt["initial_hessian"]
            actions.append({"initial_hessian": "deleted"})

        os.replace(self._input_path, os.path.join(directory, self.input_file + ".last"))
        self.qcinp.write_file(self._input_path)
        return {"errors": self.errors, "warnings": self.warnings, "actions": actions}



from custodian.utils import tracked_lru_cache


def test_cache_and_clear() -> None:
    n_calls = 0

    @tracked_lru_cache
    def some_func(x):
        nonlocal n_calls
        n_calls += 1
        return x

    assert some_func(1) == 1
    assert n_calls == 1
    assert some_func(2) == 2
    assert n_calls == 2
    assert some_func(1) == 1
    assert n_calls == 2

    assert len(tracked_lru_cache.cached_functions) == 1

    tracked_lru_cache.tracked_cache_clear()

    assert len(tracked_lru_cache.cached_functions) == 0

    assert some_func(1) == 1
    assert n_calls == 3



"""
This module implements error handlers for Nwchem runs. Currently tested only
for B3LYP DFT jobs.
"""

import os

from pymatgen.io.nwchem import NwInput, NwOutput

from custodian.ansible.interpreter import Modder
from custodian.custodian import ErrorHandler
from custodian.utils import backup


class NwchemErrorHandler(ErrorHandler):
    """
    Error handler for Nwchem Jobs. Currently tested only for B3LYP DFT jobs
    generated by pymatgen.
    """

    def __init__(self, output_filename="mol.nwout") -> None:
        """Initialize with an output file name.

        Args:
            output_filename (str): This is the file where the stdout for nwchem
                is being redirected. The error messages that are checked are
                present in the stdout. Defaults to "mol.nwout", which is the
                default redirect used by :class:`custodian.nwchem.jobs
                .NwchemJob`.
        """
        self.output_filename = output_filename

    def check(self, directory="./"):
        """Check for errors."""
        out = NwOutput(os.path.join(directory, self.output_filename))
        self.errors = []
        self.input_file = out.job_info["input"]
        if out.data[-1]["has_error"]:
            self.errors += out.data[-1]["errors"]
        self.errors = list(set(self.errors))
        self.ntasks = len(out.data)
        return len(self.errors) > 0

    def correct(self, directory="./"):
        """Correct errors."""
        backup("*.nw*", directory=directory)
        actions = []
        nwi = NwInput.from_file(os.path.join(directory, self.input_file))
        for err in self.errors:
            if err == "autoz error":
                action = {"_set": {"geometry_options": ["units", "angstroms", "noautoz"]}}
                actions.append(action)
            elif err == "Bad convergence":
                task = nwi.tasks[self.ntasks - 1]
                if "cgmin" in task.theory_directives:
                    nwi.tasks.pop(self.ntasks - 1)
                else:
                    task.theory_directives["cgmin"] = ""
                for task in nwi.tasks:
                    if task.operation.startswith("freq"):
                        # You cannot calculate hessian with cgmin.
                        task.theory_directives["nocgmin"] = ""
                action = {"_set": {"tasks": [t.as_dict() for t in nwi.tasks]}}
                actions.append(action)
            else:
                # For unimplemented errors, this should just cause the job to
                # die.
                return {"errors": self.errors, "actions": None}

        m = Modder(directory=directory)
        for action in actions:
            nwi = m.modify_object(action, nwi)
        nwi.write_file(self.input_file)
        return {"errors": self.errors, "actions": actions}

    def __str__(self) -> str:
        return "NwchemErrorHandler"



"""This module implements basic kinds of jobs for VASP runs."""

import logging
import math
import os
import shutil
import subprocess
from shutil import which

import numpy as np
import psutil
from monty.serialization import dumpfn, loadfn
from monty.shutil import decompress_dir
from pymatgen.core.structure import Structure
from pymatgen.io.vasp.inputs import Incar, Kpoints, Poscar, VaspInput
from pymatgen.io.vasp.outputs import Outcar, Vasprun

from custodian.custodian import SENTRY_DSN, Job
from custodian.utils import backup
from custodian.vasp.handlers import VASP_BACKUP_FILES
from custodian.vasp.interpreter import VaspModder

logger = logging.getLogger(__name__)


VASP_INPUT_FILES = ("INCAR", "POSCAR", "POTCAR", "KPOINTS")

VASP_OUTPUT_FILES = (
    "DOSCAR",
    "INCAR",
    "KPOINTS",
    "POSCAR",
    "PROCAR",
    "vasprun.xml",
    "CHGCAR",
    "CHG",
    "EIGENVAL",
    "OSZICAR",
    "WAVECAR",
    "CONTCAR",
    "IBZKPT",
    "OUTCAR",
)

VASP_NEB_INPUT_FILES = ("INCAR", "POTCAR", "KPOINTS")

VASP_NEB_OUTPUT_FILES = ("INCAR", "KPOINTS", "POTCAR", "vasprun.xml")

VASP_NEB_OUTPUT_SUB_FILES = (
    "CHG",
    "CHGCAR",
    "CONTCAR",
    "DOSCAR",
    "EIGENVAL",
    "IBZKPT",
    "PCDAT",
    "POSCAR",
    "REPORT",
    "PROCAR",
    "OSZICAR",
    "OUTCAR",
    "WAVECAR",
    "XDATCAR",
)


class VaspJob(Job):
    """
    A basic vasp job. Just runs whatever is in the directory. But conceivably
    can be a complex processing of inputs etc. with initialization.
    """

    def __init__(
        self,
        vasp_cmd,
        output_file="vasp.out",
        stderr_file="std_err.txt",
        suffix="",
        final=True,
        backup=True,
        auto_npar=False,
        auto_gamma=True,
        settings_override=None,
        gamma_vasp_cmd=None,
        copy_magmom=False,
        auto_continue=False,
        update_incar=False,
    ) -> None:
        """
        This constructor is necessarily complex due to the need for
        flexibility. For standard kinds of runs, it's often better to use one
        of the static constructors. The defaults are usually fine too.

        Args:
            vasp_cmd (str): Command to run vasp as a list of args. For example,
                if you are using mpirun, it can be something like
                ["mpirun", "pvasp.5.2.11"]
            output_file (str): Name of file to direct standard out to.
                Defaults to "vasp.out".
            stderr_file (str): Name of file to direct standard error to.
                Defaults to "std_err.txt".
            suffix (str): A suffix to be appended to the final output. E.g.,
                to rename all VASP output from say vasp.out to
                vasp.out.relax1, provide ".relax1" as the suffix.
            final (bool): Indicating whether this is the final vasp job in a
                series. Defaults to True.
            backup (bool): Whether to backup the initial input files. If True,
                the INCAR, KPOINTS, POSCAR and POTCAR will be copied with a
                ".orig" appended. Defaults to True.
            auto_npar (bool): Whether to automatically tune NPAR to be sqrt(
                number of cores) as recommended by VASP for DFT calculations.
                Generally, this results in significant speedups. Defaults to
                False. Set to False for HF, GW and RPA calculations.
            auto_gamma (bool): Whether to automatically check if run is a
                Gamma 1x1x1 run, and whether a Gamma optimized version of
                VASP exists with ".gamma" appended to the name of the VASP
                executable (typical setup in many systems). If so, run the
                gamma optimized version of VASP instead of regular VASP. You
                can also specify the gamma vasp command using the
                gamma_vasp_cmd argument if the command is named differently.
            settings_override ([dict]): An ansible style list of dict to
                override changes. For example, to set ISTART=1 for subsequent
                runs and to copy the CONTCAR to the POSCAR, you will provide::

                    [{"dict": "INCAR", "action": {"_set": {"ISTART": 1}}},
                     {"file": "CONTCAR",
                      "action": {"_file_copy": {"dest": "POSCAR"}}}]
            gamma_vasp_cmd (str): Command for gamma vasp version when
                auto_gamma is True. Should follow the list style of
                subprocess. Defaults to None, which means ".gamma" is added
                to the last argument of the standard vasp_cmd.
            copy_magmom (bool): Whether to copy the final magmom from the
                OUTCAR to the next INCAR. Useful for multi-relaxation runs
                where the CHGCAR and WAVECAR are sometimes deleted (due to
                changes in fft grid, etc.). Only applies to non-final runs.
            auto_continue (bool): Whether to automatically continue a run
                if a STOPCAR is present. This is very useful if using the
                wall-time handler which will write a read-only STOPCAR to
                prevent VASP from deleting it once it finishes
            update_incar (bool): Whether to update INCAR settings from updated settings
                in vasprun.xml. This is particularly useful in certain calculatons where VASP
                automatically sets certain parameters, e.g., NBANDS. Only parameters that are
                already present in the INCAR will be updated, i.e., no new parameters will be
                added even if they are in the final vasprun.xml. Note that settings_override take
                precedence over updated params.
        """
        self.vasp_cmd = tuple(vasp_cmd)
        self.output_file = output_file
        self.stderr_file = stderr_file
        self.final = final
        self.backup = backup
        self.suffix = suffix
        self.settings_override = settings_override
        self.auto_npar = auto_npar
        self.auto_gamma = auto_gamma
        self.gamma_vasp_cmd = tuple(gamma_vasp_cmd) if gamma_vasp_cmd else None
        self.copy_magmom = copy_magmom
        self.auto_continue = auto_continue
        self.update_incar = update_incar

        if SENTRY_DSN:
            # if using Sentry logging, add specific VASP executable to scope
            from sentry_sdk import configure_scope

            with configure_scope() as scope:
                try:
                    if isinstance(vasp_cmd, str):
                        vasp_path = which(vasp_cmd.split(" ")[-1])
                    elif isinstance(vasp_cmd, list):
                        vasp_path = which(vasp_cmd[-1])
                    scope.set_tag("vasp_path", vasp_path)  # type: ignore
                    scope.set_tag("vasp_cmd", vasp_cmd)
                except Exception:
                    logger.exception(f"Failed to detect VASP path: {vasp_cmd}")
                    scope.set_tag("vasp_cmd", vasp_cmd)

    def setup(self, directory="./") -> None:
        """
        Performs initial setup for VaspJob, including overriding any settings
        and backing up.
        """
        decompress_dir(directory)

        if self.backup:
            for file in VASP_INPUT_FILES:
                try:
                    shutil.copy(os.path.join(directory, file), os.path.join(directory, f"{file}.orig"))
                except FileNotFoundError:  # handle the situation when there is no KPOINTS file
                    if file == "KPOINTS":
                        pass

        if self.auto_npar:
            try:
                incar = Incar.from_file(os.path.join(directory, "INCAR"))
                # Only optimized NPAR for non-HF and non-RPA calculations.
                if not (incar.get("LHFCALC") or incar.get("LRPA") or incar.get("LEPSILON")):
                    if incar.get("IBRION") in {5, 6, 7, 8}:
                        # NPAR should not be set for Hessian matrix
                        # calculations, whether in DFPT or otherwise.
                        del incar["NPAR"]
                    else:
                        import multiprocessing

                        # try sge environment variable first
                        # (since multiprocessing counts cores on the current
                        # machine only)
                        ncores = os.environ.get("NSLOTS") or multiprocessing.cpu_count()
                        ncores = int(ncores)
                        for npar in range(int(math.sqrt(ncores)), ncores):
                            if ncores % npar == 0:
                                incar["NPAR"] = npar
                                break
                    incar.write_file(os.path.join(directory, "INCAR"))
            except Exception:
                pass

        if self.auto_continue:
            if os.path.isfile(os.path.join(directory, "continue.json")):
                actions = loadfn(os.path.join(directory, "continue.json")).get("actions")
                logger.info(f"Continuing previous VaspJob. Actions: {actions}")
                backup(VASP_BACKUP_FILES, prefix="prev_run", directory=directory)
                VaspModder(directory=directory).apply_actions(actions)

            else:
                # Default functionality is to copy CONTCAR to POSCAR and set
                # ISTART to 1 in the INCAR, but other actions can be specified
                if self.auto_continue is True:
                    actions = [
                        {
                            "file": "CONTCAR",
                            "action": {"_file_copy": {"dest": "POSCAR"}},
                        },
                        {"dict": "INCAR", "action": {"_set": {"ISTART": 1}}},
                    ]
                else:
                    actions = self.auto_continue
                dumpfn({"actions": actions}, os.path.join(directory, "continue.json"))

        if self.update_incar:
            try:
                vasprun = Vasprun(os.path.join(directory, "vasprun.xml"))
                params = vasprun.parameters
                incar = Incar.from_file(os.path.join(directory, "INCAR"))
                for k, v in incar.items():
                    incar[k] = params.get(k, v)
                incar.write_file(os.path.join(directory, "INCAR"))
            except Exception as ex:
                logger.error(f"Unable to update INCAR with params from vasprun.xml. {ex}")

        if self.settings_override is not None:
            VaspModder(directory=directory).apply_actions(self.settings_override)

    def run(self, directory="./"):
        """
        Perform the actual VASP run.

        Returns:
            (subprocess.Popen) Used for monitoring.
        """
        cmd = list(self.vasp_cmd)
        if self.auto_gamma:
            vi = VaspInput.from_directory(directory)
            if _gamma_point_only_check(vi):
                if self.gamma_vasp_cmd is not None and which(self.gamma_vasp_cmd[-1]):  # pylint: disable=E1136
                    cmd = self.gamma_vasp_cmd
                elif which(cmd[-1] + ".gamma"):
                    cmd[-1] += ".gamma"
        logger.info(f"Running {' '.join(cmd)}")
        with (
            open(os.path.join(directory, self.output_file), "w") as f_std,
            open(os.path.join(directory, self.stderr_file), "w", buffering=1) as f_err,
        ):
            # use line buffering for stderr
            return subprocess.Popen(cmd, cwd=directory, stdout=f_std, stderr=f_err, start_new_session=True)
            # pylint: disable=R1732

    def postprocess(self, directory="./") -> None:
        """
        Postprocessing includes renaming and gzipping where necessary.
        Also copies the magmom to the incar if necessary.
        """
        for file in (*VASP_OUTPUT_FILES, self.output_file):
            file = os.path.join(directory, file)
            if os.path.isfile(file):
                if self.final and self.suffix != "":
                    shutil.move(file, f"{file}{self.suffix}")
                elif self.suffix != "":
                    shutil.copy(file, f"{file}{self.suffix}")

        if self.copy_magmom and not self.final:
            try:
                outcar = Outcar(os.path.join(directory, "OUTCAR"))
                magmom = [m["tot"] for m in outcar.magnetization]
                incar = Incar.from_file(os.path.join(directory, "INCAR"))
                incar["MAGMOM"] = magmom
                incar.write_file(os.path.join(directory, "INCAR"))
            except Exception:
                logger.error("MAGMOM copy from OUTCAR to INCAR failed")

        # Remove continuation so if a subsequent job is run in
        # the same directory, will not restart this job.
        if os.path.isfile(os.path.join(directory, "continue.json")):
            os.remove(os.path.join(directory, "continue.json"))

    @classmethod
    def double_relaxation_run(
        cls,
        vasp_cmd,
        auto_npar=True,
        ediffg=-0.05,
        half_kpts_first_relax=False,
        auto_continue=False,
        directory="./",
    ):
        """
        Returns a list of two jobs corresponding to an AFLOW style double
        relaxation run.

        Args:
            vasp_cmd (str): Command to run vasp as a list of args. For example,
                if you are using mpirun, it can be something like
                ["mpirun", "pvasp.5.2.11"]
            auto_npar (bool): Whether to automatically tune NPAR to be sqrt(
                number of cores) as recommended by VASP for DFT calculations.
                Generally, this results in significant speedups. Defaults to
                True. Set to False for HF, GW and RPA calculations.
            ediffg (float): Force convergence criteria for subsequent runs (
                ignored for the initial run.)
            half_kpts_first_relax (bool): Whether to halve the kpoint grid
                for the first relaxation. Speeds up difficult convergence
                considerably. Defaults to False.
            auto_continue (bool): Whether to automatically continue a run
                if a STOPCAR is present. This is very useful if using the
                wall-time handler which will write a read-only STOPCAR to
                prevent VASP from deleting it once it finishes. Defaults to
                False.
            directory (str): Directory where the job was run. Defaults to './'.

        Returns:
            List of two jobs corresponding to an AFLOW style run.
        """
        incar_update = {"ISTART": 1}
        if ediffg:
            incar_update["EDIFFG"] = ediffg
        settings_overide_1 = None
        settings_overide_2 = [
            {"dict": "INCAR", "action": {"_set": incar_update}},
            {"file": "CONTCAR", "action": {"_file_copy": {"dest": "POSCAR"}}},
        ]
        if (
            half_kpts_first_relax
            and os.path.isfile(os.path.join(directory, "KPOINTS"))
            and os.path.isfile(os.path.join(directory, "POSCAR"))
        ):
            kpts = Kpoints.from_file(os.path.join(directory, "KPOINTS"))
            orig_kpts_dict = kpts.as_dict()
            # lattice vectors with length < 8 will get >1 KPOINT
            kpts.kpts = np.round(np.maximum(np.array(kpts.kpts) / 2, 1)).astype(int).tolist()
            low_kpts_dict = kpts.as_dict()
            settings_overide_1 = [{"dict": "KPOINTS", "action": {"_set": low_kpts_dict}}]
            settings_overide_2.append({"dict": "KPOINTS", "action": {"_set": orig_kpts_dict}})

        return [
            VaspJob(
                vasp_cmd,
                final=False,
                suffix=".relax1",
                auto_npar=auto_npar,
                auto_continue=auto_continue,
                settings_override=settings_overide_1,
            ),
            VaspJob(
                vasp_cmd,
                final=True,
                backup=False,
                suffix=".relax2",
                auto_npar=auto_npar,
                auto_continue=auto_continue,
                settings_override=settings_overide_2,
            ),
        ]

    @classmethod
    def metagga_opt_run(
        cls,
        vasp_cmd,
        auto_npar=True,
        ediffg=-0.05,
        half_kpts_first_relax=False,
        auto_continue=False,
        directory="./",
    ):
        """
        Returns a list of three jobs to perform an optimization for any
        metaGGA functional. There is an initial calculation of the
        GGA wavefunction which is fed into the initial metaGGA optimization
        to precondition the electronic structure optimizer. The metaGGA
        optimization is performed using the double relaxation scheme.
        """
        incar = Incar.from_file(os.path.join(directory, "INCAR"))
        # Defaults to using the SCAN metaGGA
        metaGGA = incar.get("METAGGA", "SCAN")

        # Pre optimize WAVECAR and structure using regular GGA
        new_settings = {"METAGGA": None, "LWAVE": True, "NSW": 0}
        pre_opt_settings = [{"dict": "INCAR", "action": {"_set": new_settings}}]
        jobs = [
            VaspJob(
                vasp_cmd,
                auto_npar=auto_npar,
                final=False,
                suffix=".precondition",
                settings_override=pre_opt_settings,
            )
        ]

        # Finish with regular double relaxation style run using SCAN
        jobs += VaspJob.double_relaxation_run(
            vasp_cmd,
            auto_npar=auto_npar,
            ediffg=ediffg,
            half_kpts_first_relax=half_kpts_first_relax,
            directory=directory,
        )

        # Ensure the first relaxation doesn't overwrite the original inputs
        jobs[1].backup = False

        # Update double_relaxation job to start from pre-optimized run
        post_opt_settings = [
            {
                "dict": "INCAR",
                "action": {
                    "_set": {
                        "METAGGA": metaGGA,
                        "ISTART": 1,
                        "NSW": incar.get("NSW", 99),
                        "LWAVE": incar.get("LWAVE", False),
                    }
                },
            },
            {"file": "CONTCAR", "action": {"_file_copy": {"dest": "POSCAR"}}},
        ]
        if jobs[1].settings_override:
            post_opt_settings = jobs[1].settings_override + post_opt_settings
        jobs[1].settings_override = post_opt_settings

        return jobs

    @classmethod
    def full_opt_run(
        cls,
        vasp_cmd,
        vol_change_tol=0.02,
        max_steps=10,
        ediffg=-0.05,
        half_kpts_first_relax=False,
        directory="./",
        **vasp_job_kwargs,
    ):
        r"""
        Returns a generator of jobs for a full optimization run. Basically,
        this runs an infinite series of geometry optimization jobs until the
        % vol change in a particular optimization is less than vol_change_tol.

        Args:
            vasp_cmd (str): Command to run vasp as a list of args. For example,
                if you are using mpirun, it can be something like
                ["mpirun", "pvasp.5.2.11"]
            vol_change_tol (float): The tolerance at which to stop a run.
                Defaults to 0.05, i.e., 5%.
            max_steps (int): The maximum number of runs. Defaults to 10 (
                highly unlikely that this limit is ever reached).
            ediffg (float): Force convergence criteria for subsequent runs (
                ignored for the initial run.)
            half_kpts_first_relax (bool): Whether to halve the kpoint grid
                for the first relaxation. Speeds up difficult convergence
                considerably. Defaults to False.
            directory (str): Directory where the job was run. Defaults to './'.
            **vasp_job_kwargs: Passthrough kwargs to VaspJob. See
                :class:`custodian.vasp.jobs.VaspJob`.

        Returns:
            Generator of jobs.
        """
        for step in range(max_steps):
            if step == 0:
                settings = None
                backup = True
                if (
                    half_kpts_first_relax
                    and os.path.isfile(os.path.join(directory, "KPOINTS"))
                    and os.path.isfile(os.path.join(directory, "POSCAR"))
                ):
                    kpts = Kpoints.from_file(os.path.join(directory, "KPOINTS"))
                    orig_kpts_dict = kpts.as_dict()
                    kpts.kpts = np.maximum(np.array(kpts.kpts) / 2, 1).tolist()
                    low_kpts_dict = kpts.as_dict()
                    settings = [{"dict": "KPOINTS", "action": {"_set": low_kpts_dict}}]
            else:
                backup = False
                initial = Poscar.from_file(os.path.join(directory, "POSCAR")).structure
                final = Poscar.from_file(os.path.join(directory, "CONTCAR")).structure
                vol_change = (final.volume - initial.volume) / initial.volume

                logger.info(f"Vol change = {vol_change:.1%}!")
                if abs(vol_change) < vol_change_tol:
                    logger.info("Stopping optimization!")
                    break
                incar_update = {"ISTART": 1}
                if ediffg:
                    incar_update["EDIFFG"] = ediffg
                settings = [
                    {"dict": "INCAR", "action": {"_set": incar_update}},
                    {
                        "file": "CONTCAR",
                        "action": {"_file_copy": {"dest": "POSCAR"}},
                    },
                ]
                if step == 1 and half_kpts_first_relax:
                    settings.append({"dict": "KPOINTS", "action": {"_set": orig_kpts_dict}})  # type: ignore
            logger.info(f"Generating job = {step + 1}!")
            yield VaspJob(
                vasp_cmd,
                final=False,
                backup=backup,
                suffix=f".relax{step + 1}",
                settings_override=settings,
                **vasp_job_kwargs,
            )

    @classmethod
    def constrained_opt_run(
        cls,
        vasp_cmd,
        lattice_direction,
        initial_strain,
        atom_relax=True,
        max_steps=20,
        algo="bfgs",
        directory="./",
        **vasp_job_kwargs,
    ):
        r"""
        Returns a generator of jobs for a constrained optimization run. Typical
        use case is when you want to approximate a biaxial strain situation,
        e.g., you apply a defined strain to a and b directions of the lattice,
        but allows the c-direction to relax.

        Some guidelines on the use of this method:
        i.  It is recommended you do not use the Auto kpoint generation. The
            grid generated via Auto may fluctuate with changes in lattice
            param, resulting in numerical noise.
        ii. Make sure your EDIFF/EDIFFG is properly set in your INCAR. The
            optimization relies on these values to determine convergence.

        Args:
            vasp_cmd (str): Command to run vasp as a list of args. For example,
                if you are using mpirun, it can be something like
                ["mpirun", "pvasp.5.2.11"]
            lattice_direction (str): Which direction to relax. Valid values are
                "a", "b" or "c".
            initial_strain (float): An initial strain to be applied to the
                lattice_direction. This can usually be estimated as the
                negative of the strain applied in the other two directions.
                E.g., if you apply a tensile strain of 0.05 to the a and b
                directions, you can use -0.05 as a reasonable first guess for
                initial strain.
            atom_relax (bool): Whether to relax atomic positions.
            max_steps (int): The maximum number of runs. Defaults to 20 (
                highly unlikely that this limit is ever reached).
            algo (str): Algorithm to use to find minimum. Default is "bfgs",
                which is fast, but can be sensitive to numerical noise
                in energy calculations. The alternative is "bisection",
                which is more robust but can be a bit slow. The code does fall
                back on the bisection when bfgs gives a nonsensical result,
                e.g., negative lattice params.
            directory (str): Directory where the job was run. Defaults to './'.
            **vasp_job_kwargs: Passthrough kwargs to VaspJob. See
                :class:`custodian.vasp.jobs.VaspJob`.

        Returns:
            Generator of jobs. At the end of the run, an "EOS.txt" is written
            which provides a quick look at the E vs lattice parameter.
        """
        nsw = 99 if atom_relax else 0

        incar = Incar.from_file(os.path.join(directory, "INCAR"))

        # Set the energy convergence criteria as the EDIFFG (if present) or
        # 10 x EDIFF (which itself defaults to 1e-4 if not present).
        e_tol = incar["EDIFFG"] if incar.get("EDIFFG") and incar.get("EDIFFG") > 0 else incar.get("EDIFF", 0.0001) * 10

        if lattice_direction == "a":
            lattice_index = 0
        elif lattice_direction == "b":
            lattice_index = 1
        else:
            lattice_index = 2

        energies = {}

        for step in range(max_steps):
            if step == 0:
                settings = [{"dict": "INCAR", "action": {"_set": {"ISIF": 2, "NSW": nsw}}}]
                structure = Poscar.from_file(os.path.join(directory, "POSCAR")).structure
                x = structure.lattice.abc[lattice_index]
                backup = True
            else:
                backup = False
                v = Vasprun("vasprun.xml")
                structure = v.final_structure
                energy = v.final_energy
                lattice = structure.lattice

                x = lattice.abc[lattice_index]

                energies[x] = energy

                if step == 1:
                    x *= 1 + initial_strain
                else:
                    # Sort the lattice parameter by energies.
                    min_x = min(energies, key=lambda e: energies[e])
                    sorted_x = sorted(energies)
                    ind = sorted_x.index(min_x)
                    if ind == 0:
                        other = ind + 1
                    elif ind == len(sorted_x) - 1:
                        other = ind - 1
                    else:
                        other = ind + 1 if energies[sorted_x[ind + 1]] < energies[sorted_x[ind - 1]] else ind - 1
                    if abs(energies[min_x] - energies[sorted_x[other]]) < e_tol:
                        logger.info(f"Stopping optimization! Final {lattice_direction} = {min_x}")
                        break

                    if ind == 0 and len(sorted_x) > 2:
                        # Lowest energy lies outside of range of lowest value.
                        # we decrease the lattice parameter in the next
                        # iteration to find a minimum. This applies only when
                        # there are at least 3 values.
                        x = sorted_x[0] - abs(sorted_x[1] - sorted_x[0])
                        logger.info(f"Lowest energy lies below bounds. Setting {lattice_direction} = {x}.")
                    elif ind == len(sorted_x) - 1 and len(sorted_x) > 2:
                        # Lowest energy lies outside of range of highest value.
                        # we increase the lattice parameter in the next
                        # iteration to find a minimum. This applies only when
                        # there are at least 3 values.
                        x = sorted_x[-1] + abs(sorted_x[-1] - sorted_x[-2])
                        logger.info(f"Lowest energy lies above bounds. Setting {lattice_direction} = {x}.")
                    else:
                        if algo.lower() == "bfgs" and len(sorted_x) >= 4:
                            try:
                                # If there are more than 4 data points, we will
                                # do a quadratic fit to accelerate convergence.
                                x1 = list(energies)
                                y1 = [energies[j] for j in x1]
                                z1 = np.polyfit(x1, y1, 2)
                                pp = np.poly1d(z1)
                                from scipy.optimize import minimize

                                result = minimize(pp, min_x, bounds=[(sorted_x[0], sorted_x[-1])])
                                if (not result.success) or result.x[0] < 0:
                                    raise ValueError("Negative lattice constant!")
                                x = result.x[0]
                                logger.info(f"BFGS minimized {lattice_direction} = {x}.")
                            except ValueError as ex:
                                # Fall back on bisection algo if the bfgs fails.
                                logger.info(str(ex))
                                x = (min_x + sorted_x[other]) / 2
                                logger.info(f"Falling back on bisection {lattice_direction} = {x}.")
                        else:
                            x = (min_x + sorted_x[other]) / 2
                            logger.info(f"Bisection {lattice_direction} = {x}.")

                lattice = lattice.matrix
                lattice[lattice_index] = lattice[lattice_index] / np.linalg.norm(lattice[lattice_index]) * x

                s = Structure(lattice, structure.species, structure.frac_coords)
                fname = os.path.join(directory, f"POSCAR.{x}")
                s.to(filename=fname)

                incar_update = {"ISTART": 1, "NSW": nsw, "ISIF": 2}

                settings = [
                    {"dict": "INCAR", "action": {"_set": incar_update}},
                    {"file": fname, "action": {"_file_copy": {"dest": "POSCAR"}}},
                ]

            logger.info(f"Generating job = {step + 1} with parameter {x}!")
            yield VaspJob(
                vasp_cmd,
                final=False,
                backup=backup,
                suffix=f".static.{x}",
                settings_override=settings,
                **vasp_job_kwargs,
            )

        with open(os.path.join(directory, "EOS.txt"), "w") as file:
            file.write(f"# {lattice_direction} energy\n")
            for key in sorted(energies):
                file.write(f"{key} {energies[key]}\n")

    def terminate(self, directory="./") -> None:
        """
        Kill all VASP processes associated with the current job.
        This is done by looping over all processes and selecting the ones
        that contain "vasp" as well as access files (vasprun.xml in particular)
        in the custodian working directory.
        There is also a safety that kills all VASP processes if none of the
        processes can be killed (This is bad if more than one VASP runs are
        simultaneously executed on the same node). However, this should never
        happen.
        """
        work_dir = directory
        logger.info(f"Killing VASP processes in {work_dir=}.")
        for proc in psutil.process_iter():
            try:
                if "vasp" in proc.name().lower():
                    open_paths = [file.path for file in proc.open_files()]
                    vasprun_path = os.path.join(work_dir, "vasprun.xml")
                    if (vasprun_path in open_paths) and psutil.pid_exists(proc.pid):
                        proc.kill()
                        return
            except (psutil.NoSuchProcess, psutil.AccessDenied) as exc:
                logger.warning(f"Exception {exc} encountered while killing VASP.")
                continue

        logger.warning(
            f"Killing VASP processes in {work_dir=} failed with subprocess.Popen.terminate(). Resorting to 'killall'."
        )
        cmds = self.vasp_cmd
        if self.gamma_vasp_cmd:
            cmds += self.gamma_vasp_cmd
        for cmd in cmds:
            if "vasp" in cmd:
                subprocess.run(["killall", f"{cmd}"], check=False)


class VaspNEBJob(VaspJob):
    """
    A NEB vasp job, especially for CI-NEB running at PBS clusters.
    The class is added for the purpose of handling a different folder
    arrangement in NEB calculation.
    """

    def __init__(
        self,
        vasp_cmd,
        output_file="neb_vasp.out",
        stderr_file="neb_std_err.txt",
        suffix="",
        final=True,
        backup=True,
        auto_npar=True,
        half_kpts=False,
        auto_gamma=True,
        auto_continue=False,
        gamma_vasp_cmd=None,
        settings_override=None,
    ) -> None:
        """
        This constructor is a simplified version of VaspJob, which satisfies
        the need for flexibility. For standard kinds of runs, it's often
        better to use one of the static constructors. The defaults are
        usually fine too.

        Args:
            vasp_cmd (str): Command to run vasp as a list of args. For example,
                if you are using mpirun, it can be something like
                ["mpirun", "pvasp.5.2.11"]
            output_file (str): Name of file to direct standard out to.
                Defaults to "vasp.out".
            stderr_file (str): Name of file to direct standard error to.
                Defaults to "std_err.txt".
            suffix (str): A suffix to be appended to the final output. E.g.,
                to rename all VASP output from say vasp.out to
                vasp.out.relax1, provide ".relax1" as the suffix.
            final (bool): Indicating whether this is the final vasp job in a
                series. Defaults to True.
            backup (bool): Whether to backup the initial input files. If True,
                the INCAR, KPOINTS, POSCAR and POTCAR will be copied with a
                ".orig" appended. Defaults to True.
            auto_npar (bool): Whether to automatically tune NPAR to be sqrt(
                number of cores) as recommended by VASP for DFT calculations.
                Generally, this results in significant speedups. Defaults to
                True. Set to False for HF, GW and RPA calculations.
            half_kpts (bool): Whether to halve the kpoint grid for NEB.
                Speeds up convergence considerably. Defaults to False.
            auto_gamma (bool): Whether to automatically check if run is a
                Gamma 1x1x1 run, and whether a Gamma optimized version of
                VASP exists with ".gamma" appended to the name of the VASP
                executable (typical setup in many systems). If so, run the
                gamma optimized version of VASP instead of regular VASP. You
                can also specify the gamma vasp command using the
                gamma_vasp_cmd argument if the command is named differently.
            auto_continue (bool): Whether to automatically continue a run
                if a STOPCAR is present. This is very useful if using the
                wall-time handler which will write a read-only STOPCAR to
                prevent VASP from deleting it once it finishes.
            gamma_vasp_cmd (str): Command for gamma vasp version when
                auto_gamma is True. Should follow the list style of
                subprocess. Defaults to None, which means ".gamma" is added
                to the last argument of the standard vasp_cmd.
            settings_override ([dict]): An ansible style list of dict to
                override changes. For example, to set ISTART=1 for subsequent
                runs and to copy the CONTCAR to the POSCAR, you will provide::

                    [{"dict": "INCAR", "action": {"_set": {"ISTART": 1}}},
                     {"file": "CONTCAR",
                      "action": {"_file_copy": {"dest": "POSCAR"}}}]
        """
        self.vasp_cmd = tuple(vasp_cmd)
        self.output_file = output_file
        self.stderr_file = stderr_file
        self.final = final
        self.backup = backup
        self.suffix = suffix
        self.auto_npar = auto_npar
        self.half_kpts = half_kpts
        self.auto_gamma = auto_gamma
        self.gamma_vasp_cmd = tuple(gamma_vasp_cmd) if gamma_vasp_cmd else None
        self.auto_continue = auto_continue
        self.settings_override = settings_override

    def setup(self, directory="./") -> None:
        """Performs initial setup for VaspNEBJob, including overriding any settings
        and backing up.
        """
        neb_dirs, neb_sub = self._get_neb_dirs(directory)

        if self.backup:
            # Back up KPOINTS, INCAR, POTCAR
            for file in VASP_NEB_INPUT_FILES:
                shutil.copy(os.path.join(directory, file), os.path.join(directory, f"{file}.orig"))
            # Back up POSCARs
            for path in neb_dirs:
                poscar = os.path.join(path, "POSCAR")
                shutil.copy(poscar, f"{poscar}.orig")

        if self.half_kpts and os.path.isfile(os.path.join(directory, "KPOINTS")):
            kpts = Kpoints.from_file(os.path.join(directory, "KPOINTS"))
            kpts.kpts = [tuple(int(nk) for nk in kv) for kv in np.maximum(np.array(kpts.kpts) / 2, 1)]  # type:ignore[misc]
            if tuple(kpts.kpts[0]) == (1, 1, 1):
                kpt_dic = kpts.as_dict()
                kpt_dic["generation_style"] = "Gamma"
                kpts = Kpoints.from_dict(kpt_dic)
            kpts.write_file(os.path.join(directory, "KPOINTS"))

        if self.auto_npar:
            try:
                incar = Incar.from_file(os.path.join(directory, "INCAR"))
                import multiprocessing

                # Try sge environment variable first
                # (since multiprocessing counts cores on the current
                # machine only)
                ncores = os.environ.get("NSLOTS") or multiprocessing.cpu_count()
                ncores = int(ncores)
                for npar in range(int(math.sqrt(ncores)), ncores):
                    if ncores % npar == 0:
                        incar["NPAR"] = npar
                        break
                incar.write_file(os.path.join(directory, "INCAR"))
            except Exception:
                pass

        if (
            self.auto_continue
            and os.path.isfile(os.path.join(directory, "STOPCAR"))
            and not os.access(os.path.join(directory, "STOPCAR"), os.W_OK)
        ):
            # Remove STOPCAR
            os.chmod(os.path.join(directory, "STOPCAR"), 0o644)
            os.remove(os.path.join(directory, "STOPCAR"))

            # Copy CONTCAR to POSCAR
            for path in neb_sub:
                contcar = os.path.join(path, "CONTCAR")
                poscar = os.path.join(path, "POSCAR")
                shutil.copy(contcar, poscar)

        if self.settings_override is not None:
            VaspModder(directory=directory).apply_actions(self.settings_override)

    def run(self, directory="./"):
        """
        Perform the actual VASP run.

        Returns:
            (subprocess.Popen) Used for monitoring.
        """
        cmd = list(self.vasp_cmd)
        if self.auto_gamma:
            vi = VaspInput.from_directory(directory)
            if _gamma_point_only_check(vi):
                if self.gamma_vasp_cmd is not None and which(self.gamma_vasp_cmd[-1]):  # pylint: disable=E1136
                    cmd = self.gamma_vasp_cmd
                elif which(cmd[-1] + ".gamma"):
                    cmd[-1] += ".gamma"
        logger.info(f"Running {' '.join(cmd)}")
        with (
            open(os.path.join(directory, self.output_file), "w") as f_std,
            open(os.path.join(directory, self.stderr_file), "w", buffering=1) as f_err,
        ):
            # Use line buffering for stderr
            return subprocess.Popen(
                cmd,
                cwd=directory,
                stdout=f_std,
                stderr=f_err,
                start_new_session=True,
            )  # pylint: disable=R1732

    def postprocess(self, directory="./") -> None:
        """Postprocessing includes renaming and gzipping where necessary."""
        # Add suffix to all sub_dir/{items}

        neb_dirs, _neb_sub = self._get_neb_dirs(directory)

        for path in neb_dirs:
            for file in VASP_NEB_OUTPUT_SUB_FILES:
                file = os.path.join(path, file)
                if os.path.isfile(file):
                    if self.final and self.suffix != "":
                        shutil.move(file, f"{file}{self.suffix}")
                    elif self.suffix != "":
                        shutil.copy(file, f"{file}{self.suffix}")

        # Add suffix to all output files
        for file in (*VASP_NEB_OUTPUT_FILES, self.output_file):
            file = os.path.join(directory, file)
            if os.path.isfile(file):
                if self.final and self.suffix != "":
                    shutil.move(file, f"{file}{self.suffix}")
                elif self.suffix != "":
                    shutil.copy(file, f"{file}{self.suffix}")

    def _get_neb_dirs(self, directory):
        neb_dirs = sorted(  # 00, 01, etc.
            path for path in os.listdir(directory) if os.path.isdir(path) and path.isdigit()
        )
        neb_sub = neb_dirs[1:-1]  # 01, 02, etc.
        return neb_dirs, neb_sub


class GenerateVaspInputJob(Job):
    """
    Generates a VASP input based on an existing directory. This is typically
    used to modify the VASP input files before the next VaspJob.
    """

    def __init__(self, input_set, contcar_only=True, **kwargs) -> None:
        """
        Args:
            input_set (str): Full path to the input set. E.g.,
                "pymatgen.io.vasp.sets.MPNonSCFSet".
            contcar_only (bool): If True (default), only CONTCAR structures
                are used as input to the input set.
            **kwargs: Additional kwargs to pass to the input set.
        """
        self.input_set = input_set
        self.contcar_only = contcar_only
        self.kwargs = kwargs

    def setup(self, directory="./") -> None:
        """Dummy setup."""

    def run(self, directory="./") -> None:
        """Run the calculation."""
        if os.path.isfile(os.path.join(directory, "CONTCAR")):
            structure = Structure.from_file(os.path.join(directory, "CONTCAR"))
        elif (not self.contcar_only) and os.path.isfile(os.path.join(directory, "POSCAR")):
            structure = Structure.from_file(os.path.join(directory, "POSCAR"))
        else:
            raise RuntimeError("No CONTCAR/POSCAR detected to generate input!")
        modname, classname = self.input_set.rsplit(".", 1)
        mod = __import__(modname, globals(), locals(), [classname], 0)
        vis = getattr(mod, classname)(structure, **self.kwargs)
        vis.write_input(directory)

    def postprocess(self, directory="./") -> None:
        """Dummy postprocess."""


def _gamma_point_only_check(vis: VaspInput) -> bool:
    """
    Check if only a single k-point is used in this calculation.

    Parameters
    -----------
    vis: VaspInput, the VASP input set for the calculation

    Returns:
    -----------
    bool: True --> use vasp_gam, False --> use vasp_std
    """
    kpts = vis["KPOINTS"]
    if (
        kpts is not None
        and kpts.style == Kpoints.supported_modes.Gamma
        and tuple(kpts.kpts[0]) == (1, 1, 1)
        and all(abs(ks) < 1.0e-6 for ks in kpts.kpts_shift)
    ):
        return True

    if (kspacing := vis["INCAR"].get("KSPACING")) is not None and vis["INCAR"].get("KGAMMA", True):
        # Get number of kpoints per axis according to the formula given by VASP:
        # https://www.vasp.at/wiki/index.php/KSPACING
        # Note that the VASP definition of the closure relation between reciprocal
        # lattice vectors b_i and direct lattice vectors a_j is not the conventional
        # b_i . a_j = 2 pi delta_ij,
        # and instead places the 2 pi factor in the formula for getting the number
        # of kpoints per axis.
        nk = [
            int(max(1, np.ceil(vis["POSCAR"].structure.lattice.reciprocal_lattice.abc[ik] / kspacing)))
            for ik in range(3)
        ]
        return np.prod(nk) == 1

    return False



"""
This module implements the main Custodian class, which manages a list of jobs
given a set of error handlers, the abstract base classes for the
ErrorHandlers and Jobs.
"""

from __future__ import annotations

import datetime
import logging
import os
import subprocess
import sys
import tarfile
import time
import warnings
from abc import abstractmethod
from ast import literal_eval
from glob import glob
from itertools import islice

from monty.json import MontyDecoder, MontyEncoder, MSONable
from monty.serialization import dumpfn, loadfn
from monty.shutil import gzip_dir
from monty.tempfile import ScratchDir

from .utils import get_execution_host_info, tracked_lru_cache

__author__ = "Shyue Ping Ong, William Davidson Richards"
__copyright__ = "Copyright 2012, The Materials Project"
__version__ = "0.2"
__maintainer__ = "Shyue Ping Ong"
__email__ = "ongsp@ucsd.edu"
__date__ = "Sep 17 2014"

logger = logging.getLogger(__name__)

# Sentry.io is a service to aggregate logs remotely, this is useful
# for Custodian to get statistics on which errors are most common.
# If you do not have a SENTRY_DSN environment variable set, or do
# not have CUSTODIAN_REPORTING_OPT_IN set to True, then
# Sentry will not be enabled.

SENTRY_DSN = os.getenv("SENTRY_DSN")
if literal_eval(os.getenv("CUSTODIAN_REPORTING_OPT_IN", "False").title()):
    # check for environment variable to automatically set SENTRY_DSN
    # will set for True, true, TRUE, etc.
    SENTRY_DSN = "https://0f7291738eb042a3af671df9fc68ae2a@sentry.io/1470881"

if SENTRY_DSN:
    import sentry_sdk

    sentry_sdk.init(dsn=SENTRY_DSN)  # pylint: disable=E0110

    with sentry_sdk.configure_scope() as scope:
        from getpass import getuser

        try:
            scope.user = {"username": getuser()}
        except Exception:
            pass

        import socket

        scope.set_tag("hostname", socket.gethostname())


class Custodian:
    """Custodian class is the manager for a list of jobs given a list of error handlers.

    The way it works is as follows:

    1. Let's say you have defined a list of jobs as [job1, job2, job3, ...] and
       you have defined a list of possible error handlers as [err1, err2, ...]
    2. Custodian will run the jobs in the order of job1, job2, ... During each
       job, custodian will monitor for errors using the handlers that have
       is_monitor == True. If an error is detected, corrective measures are
       taken and the particular job is rerun.
    3. At the end of each individual job, Custodian will run through the list of
       error handlers that have is_monitor == False. If an error is detected,
       corrective measures are taken and the particular job is rerun.

    .. attribute: max_errors

        Maximum number of errors allowed.

    .. attribute: handlers

        All error handlers (including monitors). All error handlers are used
        to check for errors at the end of a run.

    .. attribute: monitors

        Error handlers that are Monitors, i.e., handlers that monitors a job
        as it is being run.

    .. attribute: polling_time_step

        The length of time in seconds between steps in which a job is
        checked for completion.

    .. attribute: monitor_freq

        The number of polling steps before monitoring occurs. For example,
        if you have a polling_time_step of 10 seconds and a monitor_freq of
        30, this means that Custodian uses the monitors to check for errors
        every 30 x 10 = 300 seconds, i.e., 5 minutes.
    """

    LOG_FILE = "custodian.json"

    def __init__(
        self,
        handlers,
        jobs,
        validators=None,
        max_errors_per_job=None,
        max_errors=1,
        polling_time_step=10,
        monitor_freq=30,
        skip_over_errors=False,
        scratch_dir=None,
        gzipped_output=False,
        checkpoint=False,
        terminate_func=None,
        terminate_on_nonzero_returncode=True,
        directory=None,
        **kwargs,
    ) -> None:
        """Initialize a Custodian from a list of jobs and error handlers.

        Args:
            handlers ([ErrorHandler]): Error handlers. In order of priority of
                fixing.
            jobs ([Job]): Sequence of Jobs to be run. Note that this can be
                any sequence or even a generator yielding jobs.
            validators([Validator]): Validators to ensure job success
            max_errors_per_job (int): Maximum number of errors per job allowed
                before exiting. Defaults to None, which means it is set to be
                equal to max_errors..
            max_errors (int): Maximum number of total errors allowed before
                exiting. Defaults to 1.
            polling_time_step (int): The length of time in seconds between
                steps in which a job is checked for completion. Defaults to
                10 secs.
            monitor_freq (int): The number of polling steps before monitoring
                occurs. For example, if you have a polling_time_step of 10
                seconds and a monitor_freq of 30, this means that Custodian
                uses the monitors to check for errors every 30 x 10 = 300
                seconds, i.e., 5 minutes.
            skip_over_errors (bool): If set to True, custodian will skip over
                error handlers that failed (raised an Exception of some sort).
                Otherwise, custodian will simply exit on unrecoverable errors.
                The former will lead to potentially more robust performance,
                but may make it difficult to improve handlers. The latter
                will allow one to catch potentially bad error handler
                implementations. Defaults to False.
            scratch_dir (str): If this is set, any files in the current
                directory are copied to a temporary directory in a scratch
                space first before any jobs are performed, and moved back to
                the current directory upon completion of all jobs. This is
                useful in some setups where a scratch partition has much
                faster IO. To use this, set scratch_dir=root of directory you
                want to use for runs. There is no need to provide unique
                directory names; we will use python's tempfile creation
                mechanisms. A symbolic link is created during the course of
                the run in the working directory called "scratch_link" as
                users may want to sometimes check the output during the
                course of a run. If this is None (the default), the run is
                performed in the current working directory.
            gzipped_output (bool): Whether to gzip the final output to save
                space. Defaults to False.
            checkpoint (bool):  Whether to checkpoint after each successful Job.
                Checkpoints are stored as custodian.chk.#.tar.gz files. Defaults
                to False.
            terminate_func (callable): A function to be called to terminate a
                running job. If None, the default is to call Popen.terminate.
            terminate_on_nonzero_returncode (bool): If True, a non-zero return
                code on any Job will result in a termination. Defaults to True.
            directory (str): The directory to run the jobs in. Defaults to
                the current working directory.
            **kwargs: Any other kwargs are ignored. This is to allow for easy
                 subclassing and instantiation from a dict.
        """
        self.max_errors = max_errors
        self.max_errors_per_job = max_errors_per_job or max_errors
        self.jobs = jobs
        self.handlers = handlers
        self.validators = validators or []
        self.monitors = [handler for handler in handlers if handler.is_monitor]
        self.polling_time_step = polling_time_step
        self.monitor_freq = monitor_freq
        self.skip_over_errors = skip_over_errors
        self.scratch_dir = scratch_dir
        self.gzipped_output = gzipped_output
        self.checkpoint = checkpoint
        if directory is None:
            directory = os.getcwd()
        self.directory = directory
        if self.checkpoint:
            self.restart, self.run_log = Custodian._load_checkpoint(directory)
        else:
            self.restart = 0
            self.run_log = []
        self.errors_current_job = 0
        self.total_errors = 0
        self.terminate_func = terminate_func
        self.terminate_on_nonzero_returncode = terminate_on_nonzero_returncode
        self.finished = False

    @staticmethod
    def _load_checkpoint(directory):
        restart = 0
        run_log = []
        chk_pts = glob(os.path.join(directory, "custodian.chk.*.tar.gz"))
        if chk_pts:
            chk_pt = min(chk_pts, key=lambda c: int(c.split(".")[-3]))
            restart = int(chk_pt.split(".")[-3])
            logger.info(f"Loading from checkpoint file {chk_pt}...")
            with tarfile.open(chk_pt) as file:

                def is_within_directory(directory, target):
                    abs_directory = os.path.abspath(directory)
                    abs_target = os.path.abspath(target)

                    prefix = os.path.commonprefix([abs_directory, abs_target])

                    return prefix == abs_directory

                def safe_extract(tar, path=directory, members=None, *, numeric_owner=False) -> None:
                    for member in tar.getmembers():
                        member_path = os.path.join(path, member.name)
                        if not is_within_directory(path, member_path):
                            raise Exception("Attempted Path Traversal in Tar File")

                    tar.extractall(path, members, numeric_owner=numeric_owner)

                safe_extract(file, directory)
            # Log the corrections to a json file.
            run_log = loadfn(os.path.join(directory, Custodian.LOG_FILE), cls=MontyDecoder)

        return restart, run_log

    @staticmethod
    def _delete_checkpoints(directory) -> None:
        for file in glob(os.path.join(directory, "custodian.chk.*.tar.gz")):
            os.remove(file)

    @staticmethod
    def _save_checkpoint(directory, index) -> None:
        try:
            Custodian._delete_checkpoints(directory)
            n = os.path.join(directory, f"custodian.chk.{index}.tar.gz")
            with tarfile.open(n, mode="w:gz", compresslevel=3) as file:
                file.add(directory, arcname=".")
            logger.info(f"Checkpoint written to {n}")
        except Exception:
            logger.info("Checkpointing failed")
            import traceback

            logger.error(traceback.format_exc())

    @classmethod
    def from_spec(cls, spec):
        """
        Load a Custodian instance where the jobs are specified from a
        structure and a spec dict. This allows simple
        custom job sequences to be constructed quickly via a YAML file.

        Args:
            spec (dict): A dict specifying job. A sample of the dict in
                YAML format for the usual MP workflow is given as follows

                ```
                jobs:
                - jb: custodian.vasp.jobs.VaspJob
                  params:
                    final: False
                    suffix: .relax1
                - jb: custodian.vasp.jobs.VaspJob
                  params:
                    final: True
                    suffix: .relax2
                    settings_override: {"file": "CONTCAR", "action": {"_file_copy": {"dest": "POSCAR"}}
                jobs_common_params:
                  vasp_cmd: /opt/vasp
                handlers:
                - hdlr: custodian.vasp.handlers.VaspErrorHandler
                - hdlr: custodian.vasp.handlers.AliasingErrorHandler
                - hdlr: custodian.vasp.handlers.MeshSymmetryHandler
                validators:
                - vldr: custodian.vasp.validators.VasprunXMLValidator
                custodian_params:
                  scratch_dir: /tmp
                ```

                The `jobs` key is a list of jobs. Each job is
                specified via "job": <explicit path>, and all parameters are
                specified via `params` which is a dict.

                `common_params` specify a common set of parameters that are
                passed to all jobs, e.g., vasp_cmd.

        Returns:
            Custodian instance.
        """
        dec = MontyDecoder()

        def load_class(dotpath):
            mod_name, classname = dotpath.rsplit(".", 1)
            mod = __import__(mod_name, globals(), locals(), [classname], 0)
            return getattr(mod, classname)

        def process_params(dct):
            decoded = {}
            for key, val in dct.items():
                if key.startswith("$"):
                    if isinstance(val, list):
                        val = [os.path.expandvars(path) for path in val]
                    elif isinstance(val, dict):
                        val = {k2: os.path.expandvars(v2) for k2, v2 in val.items()}
                    else:
                        val = os.path.expandvars(val)
                decoded[key.strip("$")] = dec.process_decoded(val)
            return decoded

        jobs = []
        common_params = process_params(spec.get("jobs_common_params", {}))

        for dct in spec["jobs"]:
            cls_ = load_class(dct["jb"])
            params = process_params(dct.get("params", {}))
            params.update(common_params)
            jobs.append(cls_(**params))

        handlers = []
        for dct in spec.get("handlers", []):
            cls_ = load_class(dct["hdlr"])
            params = process_params(dct.get("params", {}))
            handlers.append(cls_(**params))

        validators = []
        for dct in spec.get("validators", []):
            cls_ = load_class(dct["vldr"])
            params = process_params(dct.get("params", {}))
            validators.append(cls_(**params))

        custodian_params = process_params(spec.get("custodian_params", {}))

        return cls(jobs=jobs, handlers=handlers, validators=validators, **custodian_params)

    def run(self):
        """
        Runs all jobs.

        Returns:
            All errors encountered as a list of list.
            [[error_dicts for job 1], [error_dicts for job 2], ....]

        Raises:
            ValidationError: if a job fails validation
            ReturnCodeError: if the process has a return code different from 0
            NonRecoverableError: if an unrecoverable occurs
            MaxCorrectionsPerJobError: if max_errors_per_job is reached
            MaxCorrectionsError: if max_errors is reached
            MaxCorrectionsPerHandlerError: if max_errors_per_handler is reached
        """
        with ScratchDir(
            self.scratch_dir,
            create_symbolic_link=True,
            copy_to_current_on_exit=True,
            copy_from_current_on_enter=True,
        ) as temp_dir:
            self.total_errors = 0
            start = datetime.datetime.now()
            logger.info(f"Run started at {start} in {temp_dir}.")
            v = sys.version.replace("\n", " ")
            logger.info(f"Custodian running on Python version {v}")
            host, cluster = get_execution_host_info()
            logger.info(f"Hostname: {host}, Cluster: {cluster}")

            try:
                # skip jobs until the restart
                for job_n, job in islice(enumerate(self.jobs, start=1), self.restart, None):
                    self._run_job(job_n, job)
                    # We do a dump of the run log after each job.
                    dumpfn(self.run_log, os.path.join(self.directory, Custodian.LOG_FILE), cls=MontyEncoder, indent=4)
                    # Checkpoint after each job so that we can recover from last
                    # point and remove old checkpoints
                    if self.checkpoint:
                        self.restart = job_n
                        Custodian._save_checkpoint(self.directory, job_n)
            except CustodianError as ex:
                logger.error(ex.message)
                if ex.raises:
                    raise
            finally:
                # Log the corrections to a json file.
                logger.info(f"Logging to {os.path.join(self.directory, Custodian.LOG_FILE)}...")
                dumpfn(self.run_log, os.path.join(self.directory, Custodian.LOG_FILE), cls=MontyEncoder, indent=4)
                end = datetime.datetime.now()
                logger.info(f"Run ended at {end}.")
                run_time = end - start
                logger.info(f"Run completed. Total time taken = {run_time}.")
                if self.gzipped_output:
                    gzip_dir(".")

            # Cleanup checkpoint files (if any) if run is successful.
            Custodian._delete_checkpoints(self.directory)

        return self.run_log

    def _run_job(self, job_n, job) -> None:
        """
        Runs a single job.

        Args:
            job_n: job number (1 index)
            job: Custodian job


        Raises:
            ValidationError: if a job fails validation
            ReturnCodeError: if the process has a return code different from 0
            NonRecoverableError: if an unrecoverable occurs
            MaxCorrectionsPerJobError: if max_errors_per_job is reached
            MaxCorrectionsError: if max_errors is reached
            MaxCorrectionsPerHandlerError: if max_errors_per_handler is reached
        """
        self.run_log.append(
            {
                "job": job.as_dict(),
                "corrections": [],
                "handler": None,
                "validator": None,
                "max_errors": False,
                "max_errors_per_job": False,
                "max_errors_per_handler": False,
                "nonzero_return_code": False,
            }
        )
        self.errors_current_job = 0
        # reset the counters of the number of times a correction has been
        # applied for each handler
        for handler in self.handlers:
            handler.n_applied_corrections = 0

        job.setup(self.directory)

        attempt = 0
        while self.total_errors < self.max_errors and self.errors_current_job < self.max_errors_per_job:
            attempt += 1
            logger.info(
                f"Starting job no. {job_n} ({job.name}) attempt no. {attempt}. Total errors and "
                f"errors in job thus far = {self.total_errors}, {self.errors_current_job}."
            )

            p = job.run(directory=self.directory)
            # Check for errors using the error handlers and perform
            # corrections.
            has_error = False
            zero_return_code = True

            # Choose the terminate function to run. If a terminate_func exists, this
            # should take priority, followed by Job.terminate if implemented, and finally
            # subprocess.Popen.terminate if neither of the former exist.
            terminate = self.terminate_func or job.terminate or p.terminate

            # While the job is running, we use the handlers that are
            # monitors to monitor the job.
            if isinstance(p, subprocess.Popen):
                if self.monitors:
                    n = 0
                    while True:
                        n += 1
                        time.sleep(self.polling_time_step)
                        # We poll the process p to check if it is still running.
                        # Note that the process here is not the actual calculation
                        # but whatever is used to control the execution of the
                        # calculation executable. For instance; mpirun, srun, and so on.
                        if p.poll() is not None:
                            break
                        if n % self.monitor_freq == 0:
                            # At every self.polling_time_step * self.monitor_freq seconds,
                            # we check the job for errors using handlers that are monitors.
                            # In order to properly kill a running calculation, we use
                            # the appropriate implementation of terminate.
                            has_error = self._do_check(self.monitors, terminate)
                else:
                    p.wait()
                    if self.terminate_func is not None and self.terminate_func != p.terminate:
                        self.terminate_func()
                        time.sleep(self.polling_time_step)

                zero_return_code = p.returncode == 0

            logger.info(f"{job.name}.run has completed. Checking remaining handlers")
            # Check for errors again, since in some cases non-monitor
            # handlers fix the problems detected by monitors
            # if an error has been found, not all handlers need to run
            if has_error:
                self._do_check([handler for handler in self.handlers if not handler.is_monitor])
            else:
                has_error = self._do_check(self.handlers)

            # If there are no errors detected, perform
            # postprocessing and exit.
            if not has_error:
                for validator in self.validators:
                    if validator.check(self.directory):
                        self.run_log[-1]["validator"] = validator
                        msg = f"Validation failed: {type(validator).__name__}"
                        raise ValidationError(msg, raises=True, validator=validator)
                if not zero_return_code:
                    if self.terminate_on_nonzero_returncode:
                        self.run_log[-1]["nonzero_return_code"] = True
                        msg = f"Job return code is {p.returncode}. Terminating..."
                        logger.info(msg)
                        raise ReturnCodeError(msg, raises=True)
                    warnings.warn("subprocess returned a non-zero return code. Check outputs carefully...")
                job.postprocess(directory=self.directory)
                return

            # Check that all errors could be handled
            for corr in self.run_log[-1]["corrections"]:
                if not corr["actions"] and corr["handler"].raises_runtime_error:
                    self.run_log[-1]["handler"] = corr["handler"]
                    msg = f"Unrecoverable error for handler: {corr['handler']}"
                    raise NonRecoverableError(msg, raises=True, handler=corr["handler"])
            for corr in self.run_log[-1]["corrections"]:
                if not corr["actions"]:
                    self.run_log[-1]["handler"] = corr["handler"]
                    msg = f"Unrecoverable error for handler: {corr['handler']}"
                    raise NonRecoverableError(msg, raises=False, handler=corr["handler"])

        if self.errors_current_job >= self.max_errors_per_job:
            self.run_log[-1]["max_errors_per_job"] = True
            msg = f"Max errors per job reached: {self.max_errors_per_job}."
            logger.info(msg)
            raise MaxCorrectionsPerJobError(msg, raises=True, max_errors_per_job=self.max_errors_per_job, job=job)

        self.run_log[-1]["max_errors"] = True
        msg = f"Max errors reached: {self.max_errors}."
        logger.info(msg)
        raise MaxCorrectionsError(msg, raises=True, max_errors=self.max_errors)

    def run_interrupted(self):
        """
        Runs custodian in a interrupted mode, which sets up and
        validates jobs but doesn't run the executable.

        Returns:
            number of remaining jobs

        Raises:
            ValidationError: if a job fails validation
            ReturnCodeError: if the process has a return code different from 0
            NonRecoverableError: if an unrecoverable occurs
            MaxCorrectionsPerJobError: if max_errors_per_job is reached
            MaxCorrectionsError: if max_errors is reached
            MaxCorrectionsPerHandlerError: if max_errors_per_handler is reached
        """
        start = datetime.datetime.now()
        try:
            validator = sys.version.replace("\n", " ")
            logger.info(f"Custodian started in singleshot mode at {start} in {self.directory}.")
            logger.info(f"Custodian running on Python version {validator}")

            # load run log
            if os.path.isfile(Custodian.LOG_FILE):
                self.run_log = loadfn(os.path.join(self.directory, Custodian.LOG_FILE), cls=MontyDecoder)

            if len(self.run_log) == 0:
                # starting up an initial job - setup input and quit
                job_n = 0
                job = self.jobs[job_n]
                logger.info(f"Setting up job no. 1 ({job.name}) ")
                job.setup(directory=self.directory)
                self.run_log.append({"job": job.as_dict(), "corrections": [], "job_n": job_n})
                return len(self.jobs)

            # Continuing after running calculation
            job_n = self.run_log[-1]["job_n"]
            job = self.jobs[job_n]

            # If we had to fix errors from a previous run, insert clean log
            # dict
            if len(self.run_log[-1]["corrections"]) > 0:
                logger.info(f"Reran {job.name}.run due to fixable errors")

            # check error handlers
            logger.info(f"Checking error handlers for {job.name}.run")
            if self._do_check(self.handlers):
                logger.info("Failed validation based on error handlers")
                # raise an error for an unrecoverable error
                for corr in self.run_log[-1]["corrections"]:
                    if not corr["actions"] and corr["handler"].raises_runtime_error:
                        self.run_log[-1]["handler"] = corr["handler"]
                        s = f"Unrecoverable error for handler: {corr['handler']}. Raising RuntimeError"
                        raise NonRecoverableError(s, raises=True, handler=corr["handler"])
                logger.info("Corrected input based on error handlers")
                # Return with more jobs to run if recoverable error caught
                # and corrected for
                return len(self.jobs) - job_n

            # check validators
            logger.info(f"Checking validator for {job.name}.run")
            for validator in self.validators:
                if validator.check(directory=self.directory):
                    self.run_log[-1]["validator"] = validator
                    logger.info("Failed validation based on validator")
                    s = f"Validation failed: {validator}"
                    raise ValidationError(s, raises=True, validator=validator)

            logger.info(f"Postprocessing for {job.name}.run")
            job.postprocess(directory=self.directory)

            # IF DONE WITH ALL JOBS - DELETE ALL CHECKPOINTS AND RETURN
            # VALIDATED
            if len(self.jobs) == (job_n + 1):
                self.finished = True
                return 0

            # Setup next job_n
            job_n += 1
            job = self.jobs[job_n]
            self.run_log.append({"job": job.as_dict(), "corrections": [], "job_n": job_n})
            job.setup(directory=self.directory)
            return len(self.jobs) - job_n

        except CustodianError as ex:
            logger.error(ex.message)
            if ex.raises:
                raise

        finally:
            # Log the corrections to a json file.
            logger.info(f"Logging to {Custodian.LOG_FILE}...")
            dumpfn(self.run_log, os.path.join(self.directory, Custodian.LOG_FILE), cls=MontyEncoder, indent=4)
            end = datetime.datetime.now()
            logger.info(f"Run ended at {end}.")
            run_time = end - start
            logger.info(f"Run completed. Total time taken = {run_time}.")
            if self.finished and self.gzipped_output:
                gzip_dir(self.directory)
        return None

    def _do_check(self, handlers, terminate_func=None):
        """Checks the specified handlers. Returns True iff errors caught."""
        corrections = []
        for handler in handlers:
            try:
                if handler.check(directory=self.directory):
                    if (
                        handler.max_num_corrections is not None
                        and handler.n_applied_corrections >= handler.max_num_corrections
                    ):
                        msg = f"Maximum number of corrections {handler.max_num_corrections} reached for {handler=}"
                        if handler.raise_on_max:
                            self.run_log[-1]["handler"] = handler
                            self.run_log[-1]["max_errors_per_handler"] = True
                            raise MaxCorrectionsPerHandlerError(
                                msg, raises=True, max_errors_per_handler=handler.max_num_corrections, handler=handler
                            )
                        logger.warning(f"{msg} Correction not applied.")
                        continue
                    if terminate_func is not None and handler.is_terminating:
                        logger.info("Terminating job")
                        terminate_func(directory=self.directory)
                        # make sure we don't terminate twice
                        terminate_func = None
                    dct = handler.correct(directory=self.directory)
                    logger.error(type(handler).__name__, extra=dct)
                    dct["handler"] = handler
                    corrections.append(dct)
                    handler.n_applied_corrections += 1
            except Exception:
                if not self.skip_over_errors:
                    raise
                import traceback

                logger.error(f"Bad {handler=}")
                logger.error(traceback.format_exc())
                corrections.append({"errors": [f"Bad {handler=}"], "actions": []})
        self.total_errors += len(corrections)
        self.errors_current_job += len(corrections)
        self.run_log[-1]["corrections"] += corrections
        # We do a dump of the run log after each check.
        dumpfn(self.run_log, os.path.join(self.directory, Custodian.LOG_FILE), cls=MontyEncoder, indent=4)
        # Clear all the cached values to avoid reusing them in a subsequent check
        tracked_lru_cache.tracked_cache_clear()
        return len(corrections) > 0


class Job(MSONable):
    """Abstract base class defining the interface for a Job."""

    @abstractmethod
    def setup(self, directory="./"):
        """
        This method is run before the start of a job. Allows for some
        pre-processing.
        """

    @abstractmethod
    def run(self, directory="./"):
        """
        This method perform the actual work for the job. If parallel error
        checking (monitoring) is desired, this must return a Popen process.
        """

    @abstractmethod
    def postprocess(self, directory="./"):
        """
        This method is called at the end of a job, *after* error detection.
        This allows post-processing, such as cleanup, analysis of results,
        etc.
        """

    def terminate(self, directory="./") -> None:
        """Implement termination function."""
        return

    @property
    def name(self):
        """A nice string name for the job."""
        return type(self).__name__


class ErrorHandler(MSONable):
    """Abstract base class defining the interface for an ErrorHandler."""

    is_monitor = False
    """
    This class property indicates whether the error handler is a monitor,
    i.e., a handler that monitors a job as it is running. If a
    monitor-type handler notices an error, the job will be sent a
    termination signal, the error is then corrected,
    and then the job is restarted. This is useful for catching errors
    that occur early in the run but do not cause immediate failure.
    """

    is_terminating = True
    """
    Whether this handler terminates a job upon error detection. By
    default, this is True, which means that the current Job will be
    terminated upon error detection, corrections applied,
    and restarted. In some instances, some errors may not need the job to be
    terminated or may need to wait for some other event to terminate a job.
    For example, a particular error may require a flag to be set to request
    a job to terminate gracefully once it finishes its current task. The
    handler to set the flag should be classified as is_terminating = False to
    not terminate the job.
    """

    raises_runtime_error = True
    """
    Whether this handler causes custodian to raise a runtime error if it cannot
    handle the error (i.e. if correct returns a dict with "actions":None, or
    "actions":[])
    """

    max_num_corrections: int | None = None
    raise_on_max = False
    """
    Whether corrections from this specific handler should be applied only a
    fixed maximum number of times on a single job (i.e. the counter is reset
    at the beginning of each job). If the maximum number is reached the code
    will either raise a MaxCorrectionsPerHandlerError (raise_on_max==True) or stops
    considering the correction (raise_on_max==False). If max_num_corrections
    is None this option is not considered. These options can be overridden
    as class attributes of the subclass or as customizable options setting
    an instance attribute from __init__.
    """

    @abstractmethod
    def check(self, directory="./"):
        """
        This method is called during the job (for monitors) or at the end of
        the job to check for errors.

        Returns:
            (bool) Indicating if errors are detected.
        """

    @abstractmethod
    def correct(self, directory="./"):
        """
        This method is called at the end of a job when an error is detected.
        It should perform any corrective measures relating to the detected
        error.

        Returns:
            (dict) JSON serializable dict that describes the errors and
            actions taken. E.g.
            {"errors": list_of_errors, "actions": list_of_actions_taken}.
            If this is an unfixable error, actions should be set to None.
        """

    @property
    def n_applied_corrections(self):
        """
        The number of times the handler has given a correction and this
        has been applied.

        Returns:
            (int): the number of corrections applied.
        """
        try:
            return self._num_applied_corrections
        except AttributeError:
            self._num_applied_corrections = 0
            return self._num_applied_corrections

    @n_applied_corrections.setter
    def n_applied_corrections(self, value) -> None:
        """
        Setter for the number of corrections applied.

        Args:
             value(int): the number of corrections applied
        """
        self._num_applied_corrections = value


class Validator(MSONable):
    """
    Abstract base class defining the interface for a Validator. A Validator
    differs from an ErrorHandler in that it does not correct a run and is run
    only at the end of a Job. If errors are detected by a Validator, a run is
    immediately terminated.
    """

    @abstractmethod
    def check(self, directory="./"):
        """
        This method is called at the end of a job.

        Returns:
            (bool) Indicating if errors are detected.
        """


class CustodianError(RuntimeError):
    """Exception class for Custodian errors."""

    def __init__(self, message, raises=False) -> None:
        """Initialize the error with a message.

        Args:
            message (str): Message passed to Exception
            raises (bool): Whether this should be raised outside custodian
        """
        super().__init__(message)
        self.raises = raises
        self.message = message


class ValidationError(CustodianError):
    """Error raised when a validator does not pass the check."""

    def __init__(self, message, raises, validator) -> None:
        """
        Args:
            message (str): Message passed to Exception
            raises (bool): Whether this should be raised outside custodian
            validator (Validator): Validator that caused the exception.
        """
        super().__init__(message, raises)
        self.validator = validator


class NonRecoverableError(CustodianError):
    """Error raised when a handler found an error but could not fix it."""

    def __init__(self, message, raises, handler) -> None:
        """
        Args:
            message (str): Message passed to Exception
            raises (bool): Whether this should be raised outside custodian
            handler (Handler): Handler that caused the exception.
        """
        super().__init__(message, raises)
        self.handler = handler


class ReturnCodeError(CustodianError):
    """Error raised when the process gave non zero return code."""


class MaxCorrectionsError(CustodianError):
    """Error raised when the maximum allowed number of errors is reached."""

    def __init__(self, message, raises, max_errors) -> None:
        """
        Args:
            message (str): Message passed to Exception
            raises (bool): Whether this should be raised outside custodian
            max_errors (int): the number of errors reached.
        """
        super().__init__(message, raises)
        self.max_errors = max_errors


class MaxCorrectionsPerJobError(CustodianError):
    """Error raised when the maximum allowed number of errors per job is reached."""

    def __init__(self, message, raises, max_errors_per_job, job) -> None:
        """
        Args:
            message (str): Message passed to Exception
            raises (bool): Whether this should be raised outside custodian
            max_errors_per_job (int): the number of errors per job reached
            job (Job): the job that was stopped.
        """
        super().__init__(message, raises)
        self.max_errors_per_job = max_errors_per_job
        self.job = job


class MaxCorrectionsPerHandlerError(CustodianError):
    """Error raised when the maximum allowed number of errors per handler is reached."""

    def __init__(self, message: str, raises: bool, max_errors_per_handler: int, handler: ErrorHandler) -> None:
        """
        Args:
            message (str): Message passed to Exception
            raises (bool): Whether this should be raised outside custodian
            max_errors_per_handler (int): the number of errors per job reached
            handler (Handler): the handler that caused the exception.
        """
        super().__init__(message, raises)
        self.max_errors_per_handler = max_errors_per_handler
        self.handler = handler



"""Helper functions for dealing with vasp files."""

from pymatgen.io.vasp.outputs import Outcar, Vasprun

from custodian.utils import tracked_lru_cache


@tracked_lru_cache
def load_vasprun(filepath, **vasprun_kwargs):
    """
    Load Vasprun object from file path.
    Caches the output for reuse.

    Args:
        filepath: path to the vasprun.xml file.
        **vasprun_kwargs: kwargs arguments passed to the Vasprun init.

    Returns:
        The Vasprun object
    """
    return Vasprun(filepath, **vasprun_kwargs)


@tracked_lru_cache
def load_outcar(filepath):
    """
    Load Outcar object from file path.
    Caches the output for reuse.

    Args:
        filepath: path to the OUTCAR file.

    Returns:
        The Vasprun object
    """
    return Outcar(filepath)



"""This is a master vasp running script to converging kpoints for a calculation."""

import logging

from pymatgen.io.vasp.inputs import VaspInput
from pymatgen.io.vasp.outputs import Vasprun

from custodian.custodian import Custodian
from custodian.vasp.handlers import UnconvergedErrorHandler, VaspErrorHandler
from custodian.vasp.jobs import VaspJob

FORMAT = "%(asctime)s %(message)s"
logging.basicConfig(format=FORMAT, level=logging.INFO, filename="run.log")


def get_runs(vasp_command, target=1e-3, max_steps=10, mode="linear"):
    """Generate the runs using a generator until convergence is achieved."""
    energy = 0
    vasp_input = VaspInput.from_directory(".")
    kpoints = vasp_input["KPOINTS"].kpts[0]
    for step in range(max_steps):
        m = [(kpt * (step + 1)) for kpt in kpoints] if mode == "linear" else [(kpt + 1) for kpt in kpoints]
        if step == 0:
            settings = None
            backup = True
        else:
            backup = False
            v = Vasprun("vasprun.xml")
            e_per_atom = v.final_energy / len(v.final_structure)
            ediff = abs(e_per_atom - energy)
            if ediff < target:
                logging.info(f"Converged to {ediff} eV/atom!")
                break
            energy = e_per_atom
            settings = [
                {"dict": "INCAR", "action": {"_set": {"ISTART": 1}}},
                {"dict": "KPOINTS", "action": {"_set": {"kpoints": [m]}}},
                {
                    "filename": "CONTCAR",
                    "action": {"_file_copy": {"dest": "POSCAR"}},
                },
            ]
        yield VaspJob(
            vasp_command,
            final=False,
            backup=backup,
            suffix=f".kpoints.{'x'.join(map(str, m))}",
            settings_override=settings,
        )


def do_run(args) -> None:
    """Perform the run."""
    handlers = [VaspErrorHandler(), UnconvergedErrorHandler()]
    c = Custodian(
        handlers,
        get_runs(
            vasp_command=args.command.split(),
            target=args.target,
            mode=args.mode,
            max_steps=args.max_steps,
        ),
        max_errors=10,
    )
    c.run()


def main() -> None:
    """Main method."""
    import argparse

    parser = argparse.ArgumentParser(
        description="""
    converge_kpoints perform a KPOINTS convergence. What this script will do
    is to run a particular VASP run with increasing multiples of the initial
    KPOINT grid until a target convergence in energy per atom is reached.
    For example, let's say you have vasp input files that has a k-point grid
    of 1x1x1. This script will perform sequence jobs with k-point grids of
    1x1x1, 2x2x2, 3x3x3, 4x4x4, ... until convergence is achieved. The
    default convergence criteria is 1meV/atom, but this can be set using the
    --target option.
    """,
        epilog="""Author: Shyue Ping Ong""",
    )

    parser.add_argument(
        "-c",
        "--command",
        dest="command",
        nargs="?",
        default="pvasp",
        type=str,
        help="VASP command. Defaults to pvasp. If you are using mpirun, set this to something like 'mpirun pvasp'.",
    )

    parser.add_argument(
        "-i",
        "--increment_mode",
        dest="mode",
        nargs="?",
        default="linear",
        type=str,
        choices=["linear", "inc"],
        help="Mode for increasing kpoints. In linear mode, multiples of "
        "the existing kpoints are done. E.g., 2x4x2 -> 4x8x4 -> 6x12x6. "
        "In inc mode, all KPOINTS are incremented by 1 at each stage, "
        "i.e., 2x4x2 -> 3x5x3 ->4x6x4. Note that the latter mode does "
        "not preserve KPOINTS symmetry, though it is probably less "
        "expensive.",
    )

    parser.add_argument(
        "-m",
        "--max_steps",
        dest="max_steps",
        nargs="?",
        default=10,
        type=int,
        help="The maximum number of KPOINTS increment steps. This puts an "
        "upper bound on the largest KPOINT converge grid attempted.",
    )

    parser.add_argument(
        "-t",
        "--target",
        dest="target",
        nargs="?",
        default=0.001,
        type=float,
        help="The target converge in energy per atom to achieve "
        "convergence. E.g., 1e-3 means the KPOINTS will be increased "
        "until a converged of 1meV is reached.",
    )

    args = parser.parse_args()
    do_run(args)


if __name__ == "__main__":
    main()



"""Validators for CP2K calculations."""

from __future__ import annotations

import os
from abc import abstractmethod, abstractproperty

from pymatgen.io.cp2k.outputs import Cp2kOutput

from custodian.custodian import Validator

__author__ = "Nicholas Winner"
__version__ = "1.0"
__email__ = "nwinner@berkeley.edu"
__date__ = "March 2022"


class Cp2kValidator(Validator):
    """Base validator."""

    @abstractmethod
    def check(self, directory="./"):
        """
        Check whether validation failed. Here, True means
        validation failed.
        """

    @abstractproperty
    def kill(self, directory="./"):
        """Kill the job with raise error."""

    @abstractproperty
    def exit(self, directory="./"):
        """Don't raise error, but exit the job."""

    @abstractproperty
    def no_children(self, directory="./"):
        """Job should not have children."""


class Cp2kOutputValidator(Cp2kValidator):
    """Checks that a valid cp2k output file was generated."""

    def __init__(self, output_file="cp2k.out") -> None:
        """
        Args:
            output_file (str): cp2k output file to analyze.
        """
        self.output_file = output_file
        self._check = False

    def check(self, directory="./") -> bool | None:
        """
        Check for valid output. Checks that the end of the
        program was reached, and that convergence was
        achieved.
        """
        try:
            o = Cp2kOutput(os.path.join(directory, self.output_file))
            o.ran_successfully()
            o.convergence()
            if (not o.data.get("geo_opt_converged") and not o.data.get("geo_opt_not_converged")) or o.data.get(  # type:ignore[index]
                "geo_opt_converged"
            )[-1]:
                geom = True
            else:
                geom = False
            if o.completed and o.data.get("scf_converged", [True])[-1] and geom:
                return False
            self._check = True
            return True
        except Exception:
            self._check = True
            return True

    @property
    def kill(self, directory="./") -> bool:
        """Kill the job with raise error."""
        return True

    @property
    def exit(self, directory="./") -> bool:
        """Don't raise error, but exit the job."""
        return True

    @property
    def no_children(self, directory="./") -> bool:
        """Job should not have children."""
        return True
